"""Japanese vowels dataset."""

# Author: Nathan Trouvain at 07/05/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import io
import zipfile
from urllib.request import urlopen

import numpy as np

from ._utils import _get_data_folder

SOURCE_URL = "https://archive.ics.uci.edu/static/public/128/japanese+vowels.zip"

REMOTE_FILES = {
    "DESCR": "JapaneseVowels.data.html",
    "train": "ae.train",
    "test": "ae.test",
    "train_sizes": "size_ae.train",
    "test_sizes": "size_ae.test",
}

# class labels
SPEAKERS = ["1", "2", "3", "4", "5", "6", "7", "8", "9"]

ONE_HOT_SPEAKERS = np.eye(9)


def _format_data(data, block_numbers, one_hot_encode):
    """Load and parse data from downloaded files."""

    X = []
    Y = []

    data = data.split("\n\n")[:-1]

    block_cursor = 0
    speaker_cursor = 0
    for block in data:
        if block_cursor >= block_numbers[speaker_cursor]:
            block_cursor = 0
            speaker_cursor += 1

        X.append(np.loadtxt(io.StringIO(block)))

        if one_hot_encode:
            Y.append(ONE_HOT_SPEAKERS[speaker_cursor].reshape(1, -1))
        else:
            Y.append(np.array([SPEAKERS[speaker_cursor]]).reshape(1, 1))

        block_cursor += 1

    return X, Y


def _download(data_folder):  # pragma: no cover
    """Download data from source into the reservoirpy data local directory."""

    with urlopen(SOURCE_URL) as zipresp:
        with zipfile.ZipFile(io.BytesIO(zipresp.read())) as zfile:
            zfile.extractall(data_folder)


def _repeat_target(blocks, targets):
    """Repeat target label/vector along block's time axis."""

    repeated_targets = []
    for block, target in zip(blocks, targets):
        timesteps = block.shape[0]
        target_series = np.repeat(target, timesteps, axis=0)
        repeated_targets.append(target_series)

    return repeated_targets


def japanese_vowels(
    one_hot_encode=True, repeat_targets=False, data_folder=None, reload=False
):
    """Load the Japanese vowels [16]_ dataset.

    This is a classic audio discrimination task. Nine male Japanese speakers
    pronounced the ` \\ae\\ ` vowel. The task consists in inferring the speaker
    identity from the audio recording.

    Audio recordings are series of 12 LPC cepstrum coefficient. Series contains
    between 7 and 29 timesteps. Each series (or "block") is one utterance of ` \\ae\\ `
    vowel from one speaker.

    ============================   ===============================
    Classes                                                      9
    Samples per class (training)       30 series of 7-29 timesteps
    Samples per class (testing)     29-50 series of 7-29 timesteps
    Samples total                                              640
    Dimensionality                                              12
    Features                                                  real
    ============================   ===============================

    Data is downloaded from:
    https://doi.org/10.24432/C5NS47

    Parameters
    ----------
    one_hot_encode : bool, default to True
        If True, returns class label as a one-hot encoded vector.
    repeat_targets : bool, default to False
        If True, repeat the target label or vector along the time axis of the
        corresponding sample.
    data_folder : str or Path-like object, optional
        Local destination of the downloaded data.
    reload : bool, default to False
        If True, re-download data from remote repository. Else, if a cached version
        of the dataset exists, use the cached dataset.

    Returns
    -------
    X_train, X_test, Y_train, Y_test
        Lists of arrays of shape (timesteps, features) or (timesteps, target)
        or (target,).

    References
    ----------
    .. [16] M. Kudo, J. Toyama and M. Shimbo. (1999).
           "Multidimensional Curve Classification Using Passing-Through Regions".
           Pattern Recognition Letters, Vol. 20, No. 11--13, pages 1103--1111.

    """

    data_folder = _get_data_folder(data_folder)

    complete = True
    for file_role, file_name in REMOTE_FILES.items():
        if not (data_folder / file_name).exists():
            complete = False
            break

    if reload or not complete:
        _download(data_folder)

    data_files = {}
    for file_role, file_name in REMOTE_FILES.items():
        with open(data_folder / file_name, "r") as fp:
            if file_role in ["train_sizes", "test_sizes"]:
                data = fp.read().split(" ")
                # remove empty characters and spaces
                data = [
                    int(s) for s in filter(lambda s: s not in ["", "\n", " "], data)
                ]

            else:
                data = fp.read()

        data_files[file_role] = data

    X_train, Y_train = _format_data(
        data_files["train"], data_files["train_sizes"], one_hot_encode
    )

    X_test, Y_test = _format_data(
        data_files["test"], data_files["test_sizes"], one_hot_encode
    )

    if repeat_targets:
        Y_train = _repeat_target(X_train, Y_train)
        Y_test = _repeat_target(X_test, Y_test)

    return X_train, X_test, Y_train, Y_test



# Author: Nathan Trouvain at 22/06/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from __future__ import annotations

from typing import Any, Mapping, Sequence, TypeVar, Union

import numpy as np
from scipy.sparse import issparse, sparray

global_dtype = np.float64

# Creating a real type alias (Array1D) and then using it in another alias (Timestep)
# as a str is a trick to both benefits from type checks from pyright and the like
# and also have type aliases in documentation.
# See https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#confval-autodoc_type_aliases
# See also ../docs/source/conf.py:184
Array1D = np.ndarray[tuple[int], np.dtype[np.floating]]
Array2D = np.ndarray[tuple[int, int], np.dtype[np.floating]]
Array3D = np.ndarray[tuple[int, int, int], np.dtype[np.floating]]

Timestep = Union["Array1D"]
Timeseries = Union["Array2D"]
MultiTimeseries = Union["Array3D", "Sequence[Timeseries]"]
NodeInput = Union[Timeseries, MultiTimeseries]
ModelInput = Union[NodeInput, dict[str, NodeInput]]
MappedTimestep = dict[str, Timestep]
ModelTimestep = Union[Timestep, MappedTimestep]
State = dict[str, np.ndarray]
Edge = tuple["Node", int, "Node"]
Buffer = np.ndarray[tuple[int, int], np.dtype[np.floating]]

Weights = Union[np.ndarray, sparray]
Shape = TypeVar("Shape", int, tuple[int, ...])
FeedbackBuffers = dict[Edge, Buffer]


def is_array(obj: Any) -> bool:
    return obj is not None and isinstance(obj, np.ndarray) or issparse(obj)


def is_multiseries(x: Union[NodeInput, Mapping[Any, NodeInput]]) -> bool:
    if isinstance(x, dict):
        return is_multiseries(x[list(x)[0]])
    return (isinstance(x, np.ndarray) and len(x.shape) == 3) or isinstance(x, Sequence)


def get_data_dimension(x: Union[Timestep, Timeseries, MultiTimeseries]) -> int:
    if isinstance(x, Sequence):
        if len(x) == 0:
            raise ValueError("Can't get dimension of an empty list.")
        dim = x[0].shape[-1]
    else:
        dim = x.shape[-1]  # works for both timesteps & timeseries

    return dim


def timestep_from_input(x: Union[NodeInput, Timestep]):
    if isinstance(x, Sequence):
        return np.zeros((x[0].shape[-1],))
    else:
        return np.zeros((x.shape[-1],))



# Author: Nathan Trouvain at 14/03/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest
from numpy.testing import assert_equal

from reservoirpy.nodes import Input, Output


def test_input():
    inp = Input()
    x = np.ones((10,))
    out = inp(x)
    assert_equal(out, x)
    x = np.ones((10, 10))
    out = inp.run(x)
    assert_equal(out, x)


def test_output():
    output = Output()
    x = np.ones((10,))
    out = output(x)
    assert_equal(out, x)
    x = np.ones((100, 10))
    out = output.run(x)
    assert_equal(out, x)



# Author: Nathan Trouvain at 10/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest
from numpy.testing import assert_array_equal

from reservoirpy.nodes.io import Input, Output

from ..model import Model
from ..ops import merge
from .dummy_nodes import Inverter, MinusNode, Offline, OnlineUnsupervised, PlusNode


def test_node_initialize():
    plus_node = PlusNode()
    minus_node = MinusNode()
    x = np.ones((10, 2))

    # model1 = plus_node >> minus_node
    model1 = Model([plus_node, minus_node], [(plus_node, 0, minus_node)])
    model1.initialize(x)

    # model2 = minus_node >> plus_node
    model2 = Model([plus_node, minus_node], [(minus_node, 0, plus_node)])
    model2.initialize(x)

    assert set(model1.nodes) == set(model2.nodes)

    # 2-circular graph
    with pytest.raises(RuntimeError):
        _model3 = Model(
            [plus_node, minus_node],
            [(minus_node, 0, plus_node), (plus_node, 0, minus_node)],
        )

    # 1-circular graph
    with pytest.raises(RuntimeError):
        _model4 = Model([plus_node, minus_node], [(plus_node, 0, plus_node)])


def test_multi_input():
    plus_node = PlusNode()
    x = np.ones((5,))

    # Basic multi-input
    input1, input2 = Input(name="Input1"), Input(name="Input2")
    model5 = Model(
        [input1, input2, plus_node], [(input1, 0, plus_node), (input2, 0, plus_node)]
    )
    model5.initialize({"Input1": x, "Input2": x})
    assert input1.input_dim == input2.input_dim == 5

    # multiple input but not named
    input1, input2 = Input(name="Input1"), Input()
    model5 = Model(
        [input1, input2, plus_node], [(input1, 0, plus_node), (input2, 0, plus_node)]
    )
    with pytest.raises(ValueError):
        model5.initialize({"Input1": x, "Input2": x})


def test_multi_output():
    plus_node = PlusNode()
    x = np.ones((5,))

    # Basic multi-output
    output1, output2 = Output(name="Output1"), Output(name="Output2")
    model5 = Model(
        [output1, output2, plus_node],
        [(plus_node, 0, output1), (plus_node, 0, output2)],
    )
    # res = model5(x)
    # assert output1.input_dim == output2.input_dim == 5
    # assert isinstance(res, dict)
    # assert "Output1" in res
    # assert "Output2" in res
    model5.initialize(x)

    # multiple input but not named
    output1, output2 = Output(name="Output1"), Output()
    model5 = Model(
        [output1, output2, plus_node],
        [(plus_node, 0, output1), (plus_node, 0, output2)],
    )
    with pytest.raises(ValueError):
        model5.initialize(x)


def test_complex_node_link():
    x = np.ones((10, 2))
    A = PlusNode(name="A")
    B = PlusNode(name="B")
    C = PlusNode(name="C")
    D = PlusNode(name="D")
    E = PlusNode(name="E")
    F = PlusNode(name="F")
    In = Input(name="In")
    Out = Output(name="Out")

    path1, path2 = A >> F, B >> E
    path3 = In >> [A, B, C]
    path4 = A >> B >> C >> D >> E >> F >> Out
    model = path1 & path2 & path3 & path4

    assert model.nodes == [A, F, B, E, In, C, D, Out]
    assert len(model.edges) == 11
    assert model.inputs == [In]
    assert model.outputs == [Out]
    assert len(model.named_nodes) == 8 and model.named_nodes["A"] == A
    assert not model.is_trainable
    assert not model.is_multi_input
    assert not model.is_multi_output

    model.initialize(x)

    assert len(model.parents[B]) == 2
    assert len(model.children[In]) == 3
    assert len(model.execution_order) == len(model.nodes)

    expected_dims = [2, 12, 4, 10, 2, 6, 6, 12]
    assert all([node.initialized for node in model.nodes])
    assert [node.input_dim for node in model.nodes] == expected_dims


def test_model_complex():
    from ..nodes import Reservoir, Ridge

    r1 = Reservoir(10)
    r2 = Reservoir(10)
    read1 = Ridge()
    read2 = Ridge()
    read3 = Ridge()

    model = r1 >> read1 >> r2 >> read2 >> read3
    model &= r2 & [(r1 << r2), (r1 << read1), read1] & (r2 << read2) & r2

    assert set(model.edges) == {
        (r1, 0, read1),
        (read1, 0, r2),
        (r2, 0, read2),
        (read2, 0, read3),
        (r2, 1, r1),
        (read1, 1, r1),
        (read2, 1, r2),
    }

    r1 = Reservoir(10)
    r2 = Reservoir(10)
    read1 = Ridge()
    read2 = Ridge()

    model = (r1 >> read1) & (r2 >> read2)
    model &= [r1, r2] << read1

    assert set(model.edges) == {
        (r1, 0, read1),
        (r2, 0, read2),
        (read1, 1, r1),
        (read1, 1, r2),
    }

    r1 = Reservoir(10)
    read1 = Ridge()
    model = r1 << (r1 >> read1)

    assert set(model.edges) == {(r1, 0, read1), (read1, 1, r1)}

    model = [r2] & (r1 >> read1)
    model = (r2 >> read1) & [r1]
    model = r2 & (r1 >> read1)
    model = (r2 >> read1) & r1
    model = [r2] >> (r1 >> read1)
    model = (r2 >> read1) >> [r1]
    model = r2 >> (r1 >> read1)
    model = (r2 >> read1) >> r1
    model = [r2] << (r1 >> read1)
    model = (r2 >> read1) << [r1]
    model = r2 << (r1 >> read1)
    model = (r2 >> read1) << r1


def test_model_call():
    data = np.zeros((5,))

    plus_node = PlusNode(h=2, name="Plus")
    minus_node = MinusNode(h=2, name="Minus")
    model = plus_node >> minus_node
    res = model(data)
    assert_array_equal(res, data)

    input_node = Input()
    branch1 = input_node >> plus_node
    branch2 = input_node >> minus_node
    model = branch1 & branch2
    res = model(data)
    assert model.outputs == [plus_node, minus_node]
    assert_array_equal(res["Plus"], data + 2)
    assert_array_equal(res["Minus"], data - 2)

    plus_node = PlusNode(h=2, name="Plus")
    minus_node = MinusNode(h=2, name="Minus")
    model = plus_node >> minus_node
    res = model(None)
    assert res.shape == (0,)

    plus_node1 = PlusNode(h=1, name="Plus1")
    plus_node2 = PlusNode(h=2, name="Plus2")
    minus_node = MinusNode(h=2, name="Minus")
    model = (plus_node1 >> minus_node) & (plus_node2 >> minus_node)
    res = model({"Plus1": data, "Plus2": data})
    assert res.shape == (10,)


def test_model_run():
    input_node = Input()
    plus_node = PlusNode(name="plus")
    minus_node = MinusNode(name="minus")
    branch1 = input_node >> plus_node
    branch2 = input_node >> minus_node

    model = merge(branch1, branch2)

    data = np.zeros((3, 5))
    res = model.run(data)

    expected_plus = np.ones((3, 5))
    expected_minus = -np.ones((3, 5))

    for name, arr in res.items():
        assert name in [out.name for out in model.outputs]
        if name == "plus":
            assert_array_equal(arr, expected_plus)
            assert_array_equal(arr[-1], plus_node.state["out"])
        else:
            assert_array_equal(arr, expected_minus)
            assert_array_equal(arr[-1], minus_node.state["out"])


def test_model_run_on_sequences():
    input_node = Input()
    plus_node = PlusNode(name="plus")
    minus_node = MinusNode(name="minus")
    branch1 = input_node >> plus_node
    branch2 = input_node >> minus_node

    model = branch1 & branch2

    data = np.zeros((5, 3, 5))
    res = model.run(data)

    assert set(res.keys()) == {plus_node.name, minus_node.name}
    assert len(res[plus_node.name]) == 5
    assert len(res[minus_node.name]) == 5
    assert res[plus_node.name][0].shape == (3, 5)

    input_node = Input()
    branch1 = input_node >> plus_node
    branch2 = input_node >> minus_node

    model = branch1 & branch2

    data = [np.zeros((3, 5)), np.zeros((8, 5))]
    res = model.run(data)

    assert set(res.keys()) == {plus_node.name, minus_node.name}
    assert len(res[plus_node.name]) == 2
    assert len(res[minus_node.name]) == 2
    assert res[plus_node.name][0].shape == (3, 5)
    assert res[plus_node.name][1].shape == (8, 5)


def test_offline_fit_simple_model():
    plus_node = PlusNode()
    offline_node = Offline()
    model = plus_node >> offline_node

    X = np.ones((5, 5)) * 0.5
    Y = np.ones((5, 5))

    model.fit(X, Y)

    assert offline_node.b == 25

    X = np.ones((3, 5, 5)) * 0.5
    Y = np.ones((3, 5, 5))

    model.fit(X, Y)

    assert offline_node.b == 75


def test_offline_fit_complex():
    plus_node = PlusNode()
    minus_node = MinusNode()
    offline_node1 = Offline(name="off1")
    offline_node2 = Offline(name="off2")
    model = [plus_node >> offline_node1, plus_node] >> minus_node >> offline_node2

    X = np.ones((5, 5, 5)) * 0.5
    Y_1 = np.ones((5, 5, 5))
    Y_2 = np.ones((5, 5, 10))  # after concat

    model.fit(X, y={"off1": Y_1, "off2": Y_2})

    assert plus_node.input_dim == 5
    assert plus_node.output_dim == 5
    assert minus_node.input_dim == 10
    assert minus_node.output_dim == 10
    assert offline_node1.input_dim == 5
    assert offline_node1.output_dim == 5
    assert offline_node2.input_dim == 10
    assert offline_node2.output_dim == 10

    res = model.run(X[0])

    assert res.shape == (5, 10)


def test_online_train_simple():
    plus_node = PlusNode()
    online_node = OnlineUnsupervised()
    model = plus_node >> online_node

    X = np.ones((5, 5)) * 0.5
    Y = np.ones((5, 5))

    model.partial_fit(X)

    assert online_node.b == 37.5

    model.partial_fit(X)

    assert online_node.b == 75.0

    # model.fit reinitializes variables
    model.fit(X)

    assert online_node.b == 37.5


def test_model_online():
    from reservoirpy.nodes import RLS, Output, Reservoir

    reservoir1 = Reservoir(10, name="r1")
    reservoir2 = Reservoir(10, name="r2")
    output = Output(name="out")
    readout = RLS(alpha=1e-4, name="rls")
    model = [reservoir1, reservoir2] >> readout & reservoir1 >> output

    X = np.ones((5, 5)) * 0.5
    Y = np.ones((5, 5))

    model.partial_fit({"r1": X, "r2": X}, Y)
    Wout1 = readout.Wout

    model_output = model.partial_fit({"r1": X, "r2": X}, Y)
    Wout2 = readout.Wout
    assert np.any(Wout1 != Wout2)

    assert len(model_output) == 2
    assert model_output["out"].shape == (5, 10)
    assert model_output["rls"].shape == (5, 5)


def test_model_return_states():
    off = Offline(name="offline")
    plus = PlusNode(name="plus")
    minus = MinusNode(name="minus")
    inverter = Inverter(name="inv")

    model = plus >> [minus, off >> inverter]

    X = np.ones((5, 5)) * 0.5
    Y = np.ones((5, 5))

    model.fit(X, Y)

    res = model.run(X)

    assert set(res.keys()) == {"minus", "inv"}


def test_multiinputs():
    import numpy as np

    from reservoirpy.nodes import Input, Reservoir

    source1, source2 = Input(name="s1"), Input(name="s2")
    res1, res2 = Reservoir(100, name="res1"), Reservoir(100, name="res2")
    # model = source1 >> [res1, res2] & source2 >> [res1, res2]
    model = [source1, source2] >> res1 & [source1, source2] >> res2
    outputs = model.run({"s1": np.ones((10, 5)), "s2": np.ones((10, 3))})



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from ...datasets import mackey_glass
from .. import LIF


def test_lif():
    n_timesteps = 1_000
    neurons = 100

    lif = LIF(
        units=neurons,
        inhibitory=0.0,
        sr=1.0,
        lr=0.2,
        input_scaling=1.0,
        threshold=1.0,
        rc_connectivity=1.0,
    )

    x = mackey_glass(n_timesteps=n_timesteps)
    y = lif.run(x)

    assert y.shape == (n_timesteps, neurons)
    assert_array_equal(np.sort(np.unique(y)), np.array([0.0, 1.0]))



"""Chaotic timeseries generators."""

# Author: Nathan Trouvain at 2020 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import collections
import os
from typing import Tuple, Union

import numpy as np
from joblib import Memory
from numpy.random import Generator, RandomState
from scipy.fft import fft, ifft
from scipy.integrate import solve_ivp

from .. import _TEMPDIR
from ..utils.random import rand_generator
from ._seed import get_seed

memory = Memory(os.path.join(_TEMPDIR, "datasets"), verbose=0)


def _mg_eq(xt, xtau, a=0.2, b=0.1, n=10):
    """
    Mackey-Glass time delay differential equation, at values x(t) and x(t-tau).
    """
    return -b * xt + a * xtau / (1 + xtau**n)


def _mg_rk4(xt, xtau, a, b, n, h=1.0):
    """
    Runge-Kuta method (RK4) for Mackey-Glass timeseries discretization.
    """
    k1 = h * _mg_eq(xt, xtau, a, b, n)
    k2 = h * _mg_eq(xt + 0.5 * k1, xtau, a, b, n)
    k3 = h * _mg_eq(xt + 0.5 * k2, xtau, a, b, n)
    k4 = h * _mg_eq(xt + k3, xtau, a, b, n)

    return xt + k1 / 6 + k2 / 3 + k3 / 3 + k4 / 6


def henon_map(
    n_timesteps: int,
    a: float = 1.4,
    b: float = 0.3,
    x0: Union[list, np.ndarray] = [0.0, 0.0],
    **kwargs,
) -> np.ndarray:
    """Hénon map discrete timeseries [2]_ [3]_.

    .. math::

        x(n+1) &= 1 - ax(n)^2 + y(n)\\\\
        y(n+1) &= bx(n)

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    a : float, default to 1.4
        :math:`a` parameter of the system.
    b : float, default to 0.3
        :math:`b` parameter of the system.
    x0 : array-like of shape (2,), default to [0.0, 0.0]
        Initial conditions of the system.

    Returns
    -------
    array of shape (n_timesteps, 2)
        Hénon map discrete timeseries.

    Examples
    --------
    >>> from reservoirpy.datasets import henon_map
    >>> timeseries = henon_map(2_000, a=1.4, b=0.3)
    >>> timeseries.shape
    (2000, 2)

    .. plot::

        from reservoirpy.datasets import henon_map
        plt.figure()
        timeseries = henon_map(2_000, a=1.4, b=0.3)
        plt.scatter(timeseries[:, 0], timeseries[:, 1], s=1)
        plt.xlabel("X"); plt.ylabel("Y")
        plt.show()

    References
    ----------
    .. [2] M. Hénon, ‘A two-dimensional mapping with a strange
           attractor’, Comm. Math. Phys., vol. 50, no. 1, pp. 69–77, 1976.

    .. [3] `Hénon map <https://en.wikipedia.org/wiki/H%C3%A9non_map>`_
           on Wikipedia

    """
    states = np.zeros((n_timesteps, 2))
    states[0] = np.asarray(x0)

    for i in range(1, n_timesteps):
        states[i][0] = 1 - a * states[i - 1][0] ** 2 + states[i - 1][1]
        states[i][1] = b * states[i - 1][0]

    return states


def logistic_map(n_timesteps: int, r: float = 3.9, x0: float = 0.5, **kwargs) -> np.ndarray:
    """Logistic map discrete timeseries [4]_ [5]_.

    .. math::

        x(n+1) = rx(n)(1-x(n))

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    r : float, default to 3.9
        :math:`r` parameter of the system.
    x0 : float, default to 0.5
        Initial condition of the system.

    Returns
    -------
    array of shape (n_timesteps, 1)
        Logistic map discrete timeseries.

    Examples
    --------

    >>> from reservoirpy.datasets import logistic_map
    >>> timeseries = logistic_map(n_timesteps=100)
    >>> timeseries.shape
    (100, 1)

    .. plot::

        from reservoirpy.datasets import logistic_map
        plt.figure()
        timeseries = logistic_map(100)
        plt.plot(timeseries[:100, :])
        plt.xlabel("Timesteps")
        plt.show()

    References
    ----------
    .. [4] R. M. May, ‘Simple mathematical models with very
           complicated dynamics’, Nature, vol. 261, no. 5560,
           Art. no. 5560, Jun. 1976, doi: 10.1038/261459a0.

    .. [5] `Logistic map <https://en.wikipedia.org/wiki/Logistic_map>`_
           on Wikipedia
    """
    if r > 0 and 0 < x0 < 1:
        X = np.zeros(n_timesteps)
        X[0] = x0

        for i in range(1, n_timesteps):
            X[i] = r * X[i - 1] * (1 - X[i - 1])

        return X.reshape(-1, 1)
    elif r <= 0:
        raise ValueError("r should be positive.")
    else:
        raise ValueError("Initial condition x0 should be in ]0;1[.")


def lorenz(
    n_timesteps: int,
    rho: float = 28.0,
    sigma: float = 10.0,
    beta: float = 8.0 / 3.0,
    x0: Union[list, np.ndarray] = [1.0, 1.0, 1.0],
    h: float = 0.03,
    **kwargs,
) -> np.ndarray:
    """Lorenz attractor timeseries as defined by Lorenz in 1963 [6]_ [7]_.

    .. math::

        \\frac{\\mathrm{d}x}{\\mathrm{d}t} &= \\sigma (y-x) \\\\
        \\frac{\\mathrm{d}y}{\\mathrm{d}t} &= x(\\rho - z) - y \\\\
        \\frac{\\mathrm{d}z}{\\mathrm{d}t} &= xy - \\beta z

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    rho : float, default to 28.0
        :math:`\\rho` parameter of the system.
    sigma : float, default to 10.0
        :math:`\\sigma` parameter of the system.
    beta : float, default to 8/3
        :math:`\\beta` parameter of the system.
    x0 : array-like of shape (3,), default to [1.0, 1.0, 1.0]
        Initial conditions of the system.
    h : float, default to 0.03
        Time delta between two discrete timesteps.
    **kwargs:
        Other parameters to pass to the `scipy.integrate.solve_ivp`
        solver.

    Examples
    --------

    >>> from reservoirpy.datasets import lorenz
    >>> timeseries = lorenz(10000)
    >>> timeseries.shape
    (10000, 3)

    .. plot::

        from reservoirpy.datasets import lorenz
        plt.figure()
        plt.subplot(projection="3d")
        timeseries = lorenz(10000)
        plt.plot(timeseries[:, 0], timeseries[:, 1], timeseries[:, 2], linewidth=0.1)
        plt.tight_layout()
        plt.show()

    Returns
    -------
    array of shape (n_timesteps, 3)
        Lorenz attractor timeseries.

    References
    ----------
    .. [6] E. N. Lorenz, ‘Deterministic Nonperiodic Flow’,
           Journal of the Atmospheric Sciences, vol. 20, no. 2,
           pp. 130–141, Mar. 1963,
           doi: 10.1175/1520-0469(1963)020<0130:DNF>2.0.CO;2.

    .. [7] `Lorenz system <https://en.wikipedia.org/wiki/Lorenz_system>`_
           on Wikipedia.
    """

    def lorenz_diff(t, state):
        x, y, z = state
        return sigma * (y - x), x * (rho - z) - y, x * y - beta * z

    t_max = n_timesteps * h

    t_eval = np.linspace(0.0, t_max, n_timesteps)

    sol = solve_ivp(lorenz_diff, y0=x0, t_span=(0.0, t_max), t_eval=t_eval, **kwargs)

    return sol.y.T


def mackey_glass(
    n_timesteps: int,
    tau: int = 17,
    a: float = 0.2,
    b: float = 0.1,
    n: int = 10,
    x0: float = 1.2,
    h: float = 1.0,
    seed: Union[None, int, RandomState, Generator] = None,
    history: Union[None, np.ndarray] = None,
    **kwargs,
) -> np.ndarray:
    """Mackey-Glass timeseries [8]_ [9]_, computed from the Mackey-Glass
    delayed differential equation.

    .. math::

        \\frac{x}{t} = \\frac{ax(t-\\tau)}{1+x(t-\\tau)^n} - bx(t)

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to compute.
    tau : int, default to 17
        Time delay :math:`\\tau` of Mackey-Glass equation.
        By defaults, equals to 17. Other values can
        change the chaotic behaviour of the timeseries.
    a : float, default to 0.2
        :math:`a` parameter of the equation.
    b : float, default to 0.1
        :math:`b` parameter of the equation.
    n : int, default to 10
        :math:`n` parameter of the equation.
    x0 : float, optional, default to 1.2
        Initial condition of the timeseries.
    h : float, default to 1.0
        Time delta between two discrete timesteps.
    seed : int or :py:class:`numpy.random.Generator`, optional
        Random state seed for reproducibility.
    history: array of shape (T, ), optional
        Past timesteps used to "warmup" the process.
        T must be greater than tau//h. If None, a random history is generated.


    Examples
    --------

    >>> from reservoirpy.datasets import mackey_glass
    >>> timeseries = mackey_glass(1000)
    >>> timeseries.shape
    (1000, 1)

    .. plot::

        from reservoirpy.datasets import mackey_glass
        plt.figure(figsize=(12, 3))
        timeseries = mackey_glass(500)
        plt.plot(timeseries)
        plt.show()

    Returns
    -------
    array of shape (n_timesteps, 1)
        Mackey-Glass timeseries.

    Note
    ----
        As Mackey-Glass is defined by delayed time differential equations,
        the first timesteps of the timeseries can't be initialized at 0
        (otherwise, the first steps of computation involving these
        not-computed-yet-timesteps would yield inconsistent results).
        A random number generator is therefore used to produce random
        initial timesteps based on the value of the initial condition
        passed as parameter. A default seed is hard-coded to ensure
        reproducibility in any case. It can be changed with the
        :py:func:`set_seed` function.

    References
    ----------
    .. [8] M. C. Mackey and L. Glass, ‘Oscillation and chaos in
           physiological
           control systems’, Science, vol. 197, no. 4300, pp. 287–289,
           Jul. 1977,
           doi: 10.1126/science.267326.

    .. [9] `Mackey-Glass equations
            <https://en.wikipedia.org/wiki/Mackey-Glass_equations>`_
            on Wikipedia.

    """

    history_length = int(np.floor(tau / h))

    if history is None:
        # a random state is needed as the method used to discretize
        # the timeseries needs to use randomly generated initial steps
        # based on the initial condition passed as parameter.
        if seed is None:
            seed = get_seed()

        rs = rand_generator(seed)

        # generate random first step based on the value
        # of the initial condition

        history_ = collections.deque(x0 * np.ones(history_length) + 0.2 * (rs.random(history_length) - 0.5))
    else:
        if not history_length <= len(history):
            raise ValueError(f"The given history has length of {len(history)} < tau/h" f" with tau={tau} and h={h}.")
        # use the most recent elements of the provided history
        history_ = collections.deque(history[-history_length:])

    xt = x0

    X = np.zeros(n_timesteps)

    for i in range(0, n_timesteps):
        X[i] = xt

        if tau == 0:
            xtau = 0.0
        else:
            xtau = history_.popleft()
            history_.append(xt)

        xth = _mg_rk4(xt, xtau, a=a, b=b, n=n, h=h)

        xt = xth

    return X.reshape(-1, 1)


def multiscroll(
    n_timesteps: int,
    a: float = 40.0,
    b: float = 3.0,
    c: float = 28.0,
    x0: Union[list, np.ndarray] = [-0.1, 0.5, -0.6],
    h: float = 0.01,
    **kwargs,
) -> np.ndarray:
    """Double scroll attractor timeseries [10]_ [11]_,
    a particular case of multiscroll attractor timeseries.

    .. math::

        \\frac{\\mathrm{d}x}{\\mathrm{d}t} &= a(y - x) \\\\
        \\frac{\\mathrm{d}y}{\\mathrm{d}t} &= (c - a)x - xz + cy \\\\
        \\frac{\\mathrm{d}z}{\\mathrm{d}t} &= xy - bz

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    a : float, default to 40.0
        :math:`a` parameter of the system.
    b : float, default to 3.0
        :math:`b` parameter of the system.
    c : float, default to 28.0
        :math:`c` parameter of the system.
    x0 : array-like of shape (3,), default to [-0.1, 0.5, -0.6]
        Initial conditions of the system.
    h : float, default to 0.01
        Time delta between two discrete timesteps.

    Examples
    --------

    >>> from reservoirpy.datasets import multiscroll
    >>> timeseries = multiscroll(1000)
    >>> timeseries.shape
    (1000, 3)

    .. plot::

        from reservoirpy.datasets import multiscroll
        plt.figure()
        plt.subplot(projection="3d")
        timeseries = multiscroll(10000)
        plt.plot(timeseries[:, 0], timeseries[:, 1], timeseries[:, 2], linewidth=0.1)
        plt.tight_layout()
        plt.show()

    Returns
    -------
    array of shape (n_timesteps, 3)
        Multiscroll attractor timeseries.

    References
    ----------
    .. [10] G. Chen and T. Ueta, ‘Yet another chaotic attractor’,
           Int. J. Bifurcation Chaos, vol. 09, no. 07, pp. 1465–1466,
           Jul. 1999, doi: 10.1142/S0218127499001024.

    .. [11] `Chen double scroll attractor
           <https://en.wikipedia.org/wiki/Multiscroll_attractor
           #Chen_attractor>`_
           on Wikipedia.

    """

    def multiscroll_diff(t, state):
        x, y, z = state
        dx = a * (y - x)
        dy = (c - a) * x - x * z + c * y
        dz = x * y - b * z
        return dx, dy, dz

    t_max = n_timesteps * h

    t_eval = np.linspace(0.0, t_max, n_timesteps)

    sol = solve_ivp(multiscroll_diff, y0=x0, t_span=(0.0, t_max), t_eval=t_eval, **kwargs)

    return sol.y.T


def doublescroll(
    n_timesteps: int,
    r1: float = 1.2,
    r2: float = 3.44,
    r4: float = 0.193,
    ir: float = 2 * 2.25e-5,
    beta: float = 11.6,
    x0: Union[list, np.ndarray] = [0.37926545, 0.058339, -0.08167691],
    h: float = 0.25,
    **kwargs,
) -> np.ndarray:
    """Double scroll attractor timeseries [10]_ [11]_,
    a particular case of multiscroll attractor timeseries.

    .. math::

        \\frac{\\mathrm{d}V_1}{\\mathrm{d}t} &= \\frac{V_1}{R_1} - \\frac{\\Delta V}{R_2} -
        2I_r \\sinh(\\beta\\Delta V) \\\\
        \\frac{\\mathrm{d}V_2}{\\mathrm{d}t} &= \\frac{\\Delta V}{R_2} +2I_r \\sinh(\\beta\\Delta V) - I\\\\
        \\frac{\\mathrm{d}I}{\\mathrm{d}t} &= V_2 - R_4 I

    where :math:`\\Delta V = V_1 - V_2`.

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    r1 : float, default to 1.2
        :math:`R_1` parameter of the system.
    r2 : float, default to 3.44
        :math:`R_2` parameter of the system.
    r4 : float, default to 0.193
        :math:`R_4` parameter of the system.
    ir : float, default to 2*2e.25e-5
        :math:`I_r` parameter of the system.
    beta : float, default to 11.6
        :math:`\\beta` parameter of the system.
    x0 : array-like of shape (3,), default to [0.37926545, 0.058339, -0.08167691]
        Initial conditions of the system.
    h : float, default to 0.01
        Time delta between two discrete timesteps.

    Returns
    -------
    array of shape (n_timesteps, 3)
        Multiscroll attractor timeseries.

    Examples
    --------

    >>> from reservoirpy.datasets import doublescroll
    >>> timeseries = doublescroll(1000)
    >>> timeseries.shape
    (1000, 3)

    .. plot::

        from reservoirpy.datasets import doublescroll
        plt.figure()
        plt.subplot(projection="3d")
        timeseries = doublescroll(10000)
        print(timeseries.shape)
        plt.plot(timeseries[:, 0], timeseries[:, 1], timeseries[:, 2], linewidth=0.1)
        plt.tight_layout()
        plt.show()


    References
    ----------
    .. [10] G. Chen and T. Ueta, ‘Yet another chaotic attractor’,
           Int. J. Bifurcation Chaos, vol. 09, no. 07, pp. 1465–1466,
           Jul. 1999, doi: 10.1142/S0218127499001024.

    .. [11] `Chen double scroll attractor
           <https://en.wikipedia.org/wiki/Multiscroll_attractor
           #Chen_attractor>`_
           on Wikipedia.
    """

    def doublescroll_diff(t, state):
        V1, V2, i = state

        dV = V1 - V2
        factor = (dV / r2) + ir * np.sinh(beta * dV)
        dV1 = (V1 / r1) - factor
        dV2 = factor - i
        dI = V2 - r4 * i

        return dV1, dV2, dI

    t_max = n_timesteps * h

    t_eval = np.linspace(0.0, t_max, n_timesteps)

    sol = solve_ivp(doublescroll_diff, y0=x0, t_span=(0.0, t_max), t_eval=t_eval, **kwargs)

    return sol.y.T


def rabinovich_fabrikant(
    n_timesteps: int,
    alpha: float = 1.1,
    gamma: float = 0.89,
    x0: Union[list, np.ndarray] = [-1, 0, 0.5],
    h: float = 0.05,
    **kwargs,
) -> np.ndarray:
    """Rabinovitch-Fabrikant system [12]_ [13]_ timeseries.

    .. math::

        \\frac{\\mathrm{d}x}{\\mathrm{d}t} &= y(z - 1 + x^2) + \\gamma x \\\\
        \\frac{\\mathrm{d}y}{\\mathrm{d}t} &= x(3z + 1 - x^2) + \\gamma y \\\\
        \\frac{\\mathrm{d}z}{\\mathrm{d}t} &= -2z(\\alpha + xy)

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    alpha : float, default to 1.1
        :math:`\\alpha` parameter of the system.
    gamma : float, default to 0.89
        :math:`\\gamma` parameter of the system.
    x0 : array-like of shape (3,), default to [-1, 0, 0.5]
        Initial conditions of the system.
    h : float, default to 0.05
        Time delta between two discrete timesteps.
    **kwargs:
        Other parameters to pass to the `scipy.integrate.solve_ivp`
        solver.

    Returns
    -------
    array of shape (n_timesteps, 3)
        Rabinovitch-Fabrikant system timeseries.

    Examples
    --------

    >>> from reservoirpy.datasets import rabinovich_fabrikant
    >>> timeseries = rabinovich_fabrikant(2500)
    >>> timeseries.shape
    (2500, 3)

    .. plot::

        from reservoirpy.datasets import rabinovich_fabrikant
        plt.figure()
        plt.subplot(projection="3d")
        timeseries = rabinovich_fabrikant(2500)
        print(timeseries.shape)
        plt.plot(timeseries[:, 0], timeseries[:, 1], timeseries[:, 2], linewidth=0.4)
        plt.tight_layout()
        plt.show()

    References
    ----------
    .. [12] M. I. Rabinovich and A. L. Fabrikant,
           ‘Stochastic self-modulation of waves in
           nonequilibrium media’, p. 8, 1979.

    .. [13] `Rabinovich-Fabrikant equations
           <https://en.wikipedia.org/wiki/Rabinovich%E2%80
           %93Fabrikant_equations>`_
           on Wikipedia.

    """

    def rabinovich_fabrikant_diff(t, state):
        x, y, z = state
        dx = y * (z - 1 + x**2) + gamma * x
        dy = x * (3 * z + 1 - x**2) + gamma * y
        dz = -2 * z * (alpha + x * y)
        return dx, dy, dz

    t_max = n_timesteps * h

    t_eval = np.linspace(0.0, t_max, n_timesteps)

    sol = solve_ivp(rabinovich_fabrikant_diff, y0=x0, t_span=(0.0, t_max), t_eval=t_eval, **kwargs)

    return sol.y.T


def narma(
    n_timesteps: int,
    order: int = 30,
    a1: float = 0.2,
    a2: float = 0.04,
    b: float = 1.5,
    c: float = 0.001,
    x0: Union[list, np.ndarray] = [0.0],
    seed: Union[int, RandomState] = None,
    u: np.ndarray = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Non-linear Autoregressive Moving Average (NARMA) timeseries,
    as first defined in [14]_, and as used in [15]_.

    NARMA n-th order dynamical system is defined by the recurrent relation:

    .. math::

        y[t+1] = a_1 y[t] + a_2 y[t] (\\sum_{i=0}^{n-1} y[t-i]) + b u[t-(
        n-1)]u[t] + c

    where :math:`u[t]` are sampled following a uniform distribution in
    :math:`[0, 0.5]`.

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    order: int, default to 30
        Order of the system.
    a1 : float, default to 0.2
        :math:`a_1` parameter of the system.
    a2 : float, default to 0.04
        :math:`a_2` parameter of the system.
    b : float, default to 1.5
        :math:`b` parameter of the system.
    c : float, default to 0.001
        :math:`c` parameter of the system.
    x0 : array-like of shape (init_steps,), default to [0.0]
        Initial conditions of the system.
    seed : int or :py:class:`numpy.random.Generator`, optional
        Random state seed for reproducibility.
    u : array of shape (`n_timesteps` + `order`, 1), default to None.
        Input timeseries (usually uniformly distributed).

    Returns
    -------
    tuple of arrays of shape (`n_timesteps` + `order`, 1) and (n_timesteps, 1)
        Input timeseries u and output NARMA timeseries.

    Example
    -------
    >>> import numpy as np
    >>> from reservoirpy.nodes import Reservoir, Ridge
    >>> from reservoirpy.datasets import narma
    >>> reservoir = Reservoir(100)
    >>> model = reservoir >> Ridge()
    >>> n_timesteps, order = 2000, 30
    >>> rng = np.random.default_rng(seed=2341)
    >>> u, y = narma(n_timesteps=n_timesteps, order=order)
    >>> reservoir.run(u[:order]) # warmup
    >>> model = model.fit(u[order:], y)

    .. plot::

        from reservoirpy.datasets import narma
        plt.figure(figsize=(12, 3))
        _x, timeseries = narma(100, order=30)
        plt.plot(timeseries)
        plt.show()

    References
    ----------
    .. [14] A. F. Atiya and A. G. Parlos, ‘New results on recurrent
           network training: unifying the algorithms and accelerating
           convergence,‘ in IEEE Transactions on Neural Networks,
           vol. 11, no. 3, pp. 697-709, May 2000,
           doi: 10.1109/72.846741.

    .. [15] B.Schrauwen, M. Wardermann, D. Verstraeten, J. Steil,
           D. Stroobandt, ‘Improving reservoirs using intrinsic
           plasticity‘,
           Neurocomputing, 71. 1159-1171, 2008,
           doi: 10.1016/j.neucom.2007.12.020.
    """
    if seed is None:
        seed = get_seed()
    rs = rand_generator(seed)

    y = np.zeros((n_timesteps + order, 1))

    x0 = np.atleast_2d(np.asarray(x0))
    y[: x0.shape[0], :] = x0

    if u is None:
        u = rs.uniform(0, 0.5, size=(n_timesteps + order, 1))

    for t in range(order, n_timesteps + order - 1):
        y[t + 1] = a1 * y[t] + a2 * y[t] * np.sum(y[t - order : t]) + b * u[t - order] * u[t] + c
    return u, y[order:, :]


def lorenz96(
    n_timesteps: int,
    warmup: int = 0,
    N: int = 36,
    F: float = 8.0,
    dF: float = 0.01,
    h: float = 0.01,
    x0: Union[list, np.ndarray] = None,
    **kwargs,
) -> np.ndarray:
    """Lorenz96 attractor timeseries as defined by Lorenz in 1996 [17]_.

    .. math::

        \\frac{\\mathrm{d}x_i}{\\mathrm{d} t} = (x_{i+1} - x_{i-2}) x_{i-1} - x_i + F

    where :math:`i = 1, \\dots, N` and :math:`x_{-1} = x_{N-1}`
    and :math:`x_{N+1} = x_1` and :math:`N \\geq 4`.

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    warmup : int, default to 0
        Number of timesteps to discard at the beginning of the signal, to remove
        transient states.
    N: int, default to 36
        Dimension of the system.
    F : float, default to F
        :math:`F` parameter of the system.
    dF : float, default to 0.01
        Perturbation applied to initial condition if x0 is None.
    h : float, default to 0.01
        Time delta between two discrete timesteps.
    x0 : array-like of shape (N,), default to None
        Initial conditions of the system. If None, the array is initialized to
        an array of shape (N, ) with value F, except for the first value of the
        array that takes the value F + dF.
    **kwargs:
        Other parameters to pass to the `scipy.integrate.solve_ivp`
        solver.

    Returns
    -------
    array of shape (n_timesteps - warmup, N)
        Lorenz96 timeseries.

    Examples
    --------

    >>> from reservoirpy.datasets import lorenz96
    >>> timeseries = lorenz96(1000, h=0.1, N=5)
    >>> timeseries.shape
    (1000, 5)

    .. plot::

        from reservoirpy.datasets import lorenz96
        timeseries = lorenz96(1000, h=0.1, N=5)
        plt.figure()
        plt.plot(timeseries)
        plt.show()

    References
    ----------
    .. [17] Lorenz, E. N. (1996, September).
            Predictability: A problem partly solved. In Proc.
            Seminar on predictability (Vol. 1, No. 1).
    """
    if N < 4:
        raise ValueError("N must be >= 4.")

    if x0 is None:
        x0 = F * np.ones(N)
        x0[0] = F + dF

    if len(x0) != N:
        raise ValueError(f"x0 should have shape ({N},), but have shape {np.asarray(x0).shape}")

    def lorenz96_diff(t, state):
        ds = np.zeros(N)
        for i in range(N):
            ds[i] = (state[(i + 1) % N] - state[i - 2]) * state[i - 1] - state[i] + F
        return ds

    t_max = (warmup + n_timesteps) * h

    t_eval = np.linspace(0.0, t_max * h, n_timesteps)

    sol = solve_ivp(
        lorenz96_diff,
        y0=x0,
        t_span=(0.0, t_max * h),
        t_eval=t_eval,
        **kwargs,
    )

    return sol.y.T[warmup:]


def rossler(
    n_timesteps: int,
    a: float = 0.2,
    b: float = 0.2,
    c: float = 5.7,
    x0: Union[list, np.ndarray] = [-0.1, 0.0, 0.02],
    h: float = 0.1,
    **kwargs,
) -> np.ndarray:
    """Rössler attractor timeseries [18]_.

    .. math::

        \\frac{\\mathrm{d}x}{\\mathrm{d}t} &= -y - z \\\\
        \\frac{\\mathrm{d}y}{\\mathrm{d}t} &= x + a y \\\\
        \\frac{\\mathrm{d}z}{\\mathrm{d}t} &= b + z (x - c)

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    a : float, default to 0.2
        :math:`a` parameter of the system.
    b : float, default to 0.2
        :math:`b` parameter of the system.
    c : float, default to 5.7
        :math:`c` parameter of the system.
    x0 : array-like of shape (3,), default to [-0.1, 0.0, 0.02]
        Initial conditions of the system.
    h : float, default to 0.1
        Time delta between two discrete timesteps.
    **kwargs:
        Other parameters to pass to the `scipy.integrate.solve_ivp`
        solver.

    Returns
    -------
    array of shape (n_timesteps, 3)
        Rössler attractor timeseries.

    Examples
    --------

    >>> from reservoirpy.datasets import rossler
    >>> timeseries = rossler(2000)
    >>> timeseries.shape
    (2000, 3)

    .. plot::

        from reservoirpy.datasets import rossler
        timeseries = rossler(2000)
        plt.figure()
        plt.subplot(projection="3d")
        plt.plot(timeseries[:, 0], timeseries[:, 1], timeseries[:, 2], linewidth=1)
        plt.tight_layout()
        plt.show()


    References
    ----------

    .. [18] O.E. Rössler, "An equation for continuous chaos", Physics Letters A,
            vol 57, Issue 5, Pages 397-398, ISSN 0375-9601, 1976,
            https://doi.org/10.1016/0375-9601(76)90101-8.
    """
    if len(x0) != 3:
        raise ValueError(f"x0 should have shape (3,), but have shape {np.asarray(x0).shape}")

    def rossler_diff(t, state):
        x, y, z = state
        dx = -y - z
        dy = x + a * y
        dz = b + z * (x - c)
        return dx, dy, dz

    t_max = n_timesteps * h

    t_eval = np.linspace(0.0, t_max, n_timesteps)

    sol = solve_ivp(rossler_diff, y0=x0, t_span=(0.0, t_max), t_eval=t_eval, **kwargs)

    return sol.y.T


def _kuramoto_sivashinsky_etdrk4(v, *, g, E, E2, Q, f1, f2, f3):
    """A single step of EDTRK4 to solve Kuramoto-Sivashinsky equation.

    Kassam, A. K., & Trefethen, L. N. (2005). Fourth-order time-stepping for stiff PDEs.
    SIAM Journal on Scientific Computing, 26(4), 1214-1233.
    """

    Nv = g * fft(np.real(ifft(v)) ** 2)
    a = E2 * v + Q * Nv
    Na = g * fft(np.real(ifft(a)) ** 2)
    b = E2 * v + Q * Na
    Nb = g * fft(np.real(ifft(b)) ** 2)
    c = E2 * a + Q * (2 * Nb - Nv)
    Nc = g * fft(np.real(ifft(c)) ** 2)
    v = E * v + Nv * f1 + 2 * (Na + Nb) * f2 + Nc * f3

    return v


@memory.cache
def _kuramoto_sivashinsky(n_timesteps, *, warmup, N, M, x0, h):
    # initial conditions
    v0 = fft(x0)

    # ETDRK4 scalars
    k = np.conj(np.r_[np.arange(0, N / 2), [0], np.arange(-N / 2 + 1, 0)]) / M

    L = k**2 - k**4

    E = np.exp(h * L)
    E2 = np.exp(h * L / 2)

    r = np.exp(1j * np.pi * (np.arange(1, M + 1) - 0.5) / M)
    LR = h * np.transpose(np.repeat([L], M, axis=0)) + np.repeat([r], N, axis=0)

    Q = h * np.real(np.mean((np.exp(LR / 2) - 1) / LR, axis=1))

    f1 = (-4 - LR + np.exp(LR) * (4 - 3 * LR + LR**2)) / LR**3
    f1 = h * np.real(np.mean(f1, axis=1))

    f2 = (2 + LR + np.exp(LR) * (-2 + LR)) / LR**3
    f2 = h * np.real(np.mean(f2, axis=1))

    f3 = (-4 - 3 * LR - LR**2 + np.exp(LR) * (4 - LR)) / LR**3
    f3 = h * np.real(np.mean(f3, axis=1))

    g = -0.5j * k

    # integration using ETDRK4 method
    v = np.zeros((n_timesteps, N), dtype=complex)
    v[0] = v0
    for n in range(1, n_timesteps):
        v[n] = _kuramoto_sivashinsky_etdrk4(v[n - 1], g=g, E=E, E2=E2, Q=Q, f1=f1, f2=f2, f3=f3)

    return np.real(ifft(v[warmup:]))


def kuramoto_sivashinsky(
    n_timesteps: int,
    warmup: int = 0,
    N: int = 128,
    M: float = 16,
    x0: Union[list, np.ndarray] = None,
    h: float = 0.25,
) -> np.ndarray:
    """Kuramoto-Sivashinsky oscillators [19]_ [20]_ [21]_.

    .. math::

        y_t = -yy_x - y_{xx} - y_{xxxx}, ~~ x \\in [0, 32\\pi]

    This 1D partial differential equation is solved using ETDRK4
    (Exponential Time-Differencing 4th order Runge-Kutta) method, as described in [22]_.

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    warmup : int, default to 0
        Number of timesteps to discard at the beginning of the signal, to remove
        transient states.
    N : int, default to 128
        Dimension of the system.
    M : float, default to 0.2
        Number of points for complex means. Modify behaviour of the resulting
        multivariate timeseries.
    x0 : array-like of shape (N,), default to None.
        Initial conditions of the system. If None, x0 is equal to
        :math:`\\cos (\\frac{y}{M}) * (1 + \\sin(\\frac{y}{M}))`
        with :math:`y = 2M\\pi x / N, ~~ x \\in [1, N]`.
    h : float, default to 0.25
        Time delta between two discrete timesteps.

    Returns
    -------
    array of shape (n_timesteps - warmup, N)
        Kuramoto-Sivashinsky equation solution.

    Examples
    --------

    >>> from reservoirpy.datasets import kuramoto_sivashinsky
    >>> timeseries = kuramoto_sivashinsky(1000)
    >>> print(timeseries.shape)

    .. plot::

        from reservoirpy.datasets import kuramoto_sivashinsky
        timeseries = kuramoto_sivashinsky(1000)
        plt.figure(figsize=(12, 4))
        plt.imshow(timeseries.T)
        plt.ylabel("Dimension No.")
        plt.xlabel("Time")
        plt.show()

    References
    ----------

    .. [19] Kuramoto, Y. (1978). Diffusion-Induced Chaos in Reaction Systems.
            Progress of Theoretical Physics Supplement, 64, 346–367.
            https://doi.org/10.1143/PTPS.64.346

    .. [20] Sivashinsky, G. I. (1977). Nonlinear analysis of hydrodynamic instability
            in laminar flames—I. Derivation of basic equations.
            Acta Astronautica, 4(11), 1177–1206.
            https://doi.org/10.1016/0094-5765(77)90096-0

    .. [21] Sivashinsky, G. I. (1980). On Flame Propagation Under Conditions
            of Stoichiometry. SIAM Journal on Applied Mathematics, 39(1), 67–82.
            https://doi.org/10.1137/0139007

    .. [22] Kassam, A. K., & Trefethen, L. N. (2005).
            Fourth-order time-stepping for stiff PDEs.
            SIAM Journal on Scientific Computing, 26(4), 1214-1233.
    """
    if x0 is None:
        x = 2 * M * np.pi * np.arange(1, N + 1) / N
        x0 = np.cos(x / M) * (1 + np.sin(x / M))
    else:
        if not np.asarray(x0).shape[0] == N:
            raise ValueError(
                f"Initial condition x0 should be of shape {N} (= N) but " f"has shape {np.asarray(x0).shape}"
            )
        else:
            x0 = np.asarray(x0)

    return _kuramoto_sivashinsky(n_timesteps, warmup=warmup, N=N, M=M, x0=x0, h=h)



# Author: Nathan Trouvain at 10/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from typing import Sequence

import numpy as np
import pytest

from reservoirpy.node import Node, OnlineNode, TrainableNode
from reservoirpy.type import is_multiseries


class IdentityNode(Node):
    def __init__(self, name=None):
        self.name = name
        self.initialized = False
        self.state = {}

    def _step(self, state, x):
        return {"out": x}

    def initialize(self, x, y=None):
        self.input_dim = x.shape[-1] if not isinstance(x, Sequence) else x[0].shape[-1]
        self.output_dim = self.input_dim
        self.initialized = True


class PlusNode(Node):
    def __init__(self, h=1, name=None):
        self.h = h
        self.name = name
        self.initialized = False
        self.state = {}

    def _step(self, state, x):
        return {"out": x + self.h}

    def initialize(self, x, y=None):
        self.input_dim = x.shape[-1] if not isinstance(x, Sequence) else x[0].shape[-1]
        self.output_dim = self.input_dim
        self.initialized = True


class AccumulateNode(Node):
    def __init__(self, h=0, name=None):
        self.h = h
        self.name = name
        self.initialized = False
        self.state = {}

    def _step(self, state, x):
        return {"out": x + state["out"] + self.h}

    def initialize(self, x, y=None):
        self.input_dim = x.shape[-1] if not isinstance(x, Sequence) else x[0].shape[-1]
        self.output_dim = self.input_dim
        self.state = {
            "out": np.zeros(
                self.output_dim,
            )
        }
        self.initialized = True


class MinusNode(Node):
    def __init__(self, h=1, name=None):
        self.h = h
        self.name = name
        self.initialized = False
        self.state = {}

    def _step(self, state, x):
        return {"out": x - self.h}

    def initialize(self, x, y=None):
        self.input_dim = x.shape[-1] if not isinstance(x, Sequence) else x[0].shape[-1]
        self.output_dim = self.input_dim
        self.initialized = True


class Inverter(Node):
    def __init__(self, name=None):
        self.name = name
        self.initialized = False
        self.state = {}

    def _step(self, state, x):
        return {"out": -x}

    def initialize(self, x, y=None):
        self.input_dim = x.shape[-1] if not isinstance(x, Sequence) else x[0].shape[-1]
        self.output_dim = self.input_dim
        self.initialized = True


class Offline(TrainableNode):
    def __init__(self, name=None):
        self.name = name
        self.b = 0
        self.initialized = False
        self.state = {}

    def initialize(self, x, y=None):
        self.input_dim = x.shape[-1] if not isinstance(x, Sequence) else x[0].shape[-1]
        self.output_dim = self.input_dim
        self.initialized = True

    def _step(self, state, x):
        return {"out": x + self.b}

    def fit(self, x, y, warmup=0):
        if not self.initialized:
            self.initialize(x, y)
        self.b = 0

        for el in y[warmup:]:
            self.b += np.sum(el)

        return self


class Unsupervised(TrainableNode):
    def __init__(self, name=None):
        self.name = name
        self.b = 0
        self.initialized = False
        self.state = {}

    def initialize(self, x, y=None):
        self.input_dim = x.shape[-1] if not isinstance(x, Sequence) else x[0].shape[-1]
        self.output_dim = self.input_dim
        self.initialized = True

    def _step(self, x):
        return {"out": x + self.b}

    def fit(self, x, y=None, warmup=0):
        if not self.initialized:
            self.initialize(x)

        for el in x[warmup:]:
            self.b += np.sum(el)

        return self


class OnlineUnsupervised(OnlineNode):
    def __init__(self, name=None):
        self.name = name
        self.initialized = False
        self.b = 0
        self.state = {}

    def initialize(self, x, y=None):
        self.input_dim = x.shape[-1] if not isinstance(x, Sequence) else x[0].shape[-1]
        self.output_dim = self.input_dim
        self.b = 0
        self.initialized = True

    def _step(self, state, x):
        return {"out": x + self.b}

    def _learning_step(self, x, y=None):
        self.b += np.sum(x)
        return x + self.b

    def partial_fit(self, x, y=None):
        if not self.initialized:
            self.initialize(x)

        self.b += np.sum(x)

        return self



import os
import tempfile
import uuid

from . import activationsfunc, hyper, mat_gen, nodes, observables, type
from ._version import __version__
from .esn import ESN
from .model import Model
from .node import Node
from .utils.random import set_seed

# Generate a unique temporary directory to prevent permission errors
_TEMPDIR = os.path.join(tempfile.gettempdir(), f"reservoirpy-temp-{uuid.uuid4()}")
if not os.path.exists(_TEMPDIR):  # pragma: no cover
    try:
        os.mkdir(_TEMPDIR)
    except OSError:
        _TEMPDIR = tempfile.gettempdir()

__all__ = [
    "__version__",
    "mat_gen",
    "activationsfunc",
    "observables",
    "hyper",
    "nodes",
    "Node",
    "Model",
    "set_seed",
    "ESN",
    "type",
]



import numpy as np
import pytest

from reservoirpy.utils.data_validation import (
    check_model_input,
    check_model_timestep,
    check_multiseries,
    check_node_input,
    check_timeseries,
    check_timestep,
    is_model_input,
    is_model_timestep,
    is_multiseries,
    is_node_input,
    is_timeseries,
    is_timestep,
)


def test_check_timestep():
    x = np.array([1, 2, 3])
    check_timestep(x)

    x = np.array([1, 2, 3])
    check_timestep(x, expected_dim=3)

    with pytest.raises(TypeError):
        check_timestep([1, 2, 3])

    x = np.array([[1, 2], [3, 4]])
    with pytest.raises(ValueError):
        check_timestep(x)

    x = np.array([1, 2, 3])
    with pytest.raises(ValueError):
        check_timestep(x, expected_dim=5)


def test_is_timestep():
    x = np.array([1, 2, 3])
    assert is_timestep(x) is True

    x = np.array([[1, 2], [3, 4]])
    assert is_timestep(x) is False


def test_check_timeseries():
    x = np.array([[1, 2], [3, 4], [5, 6]])
    check_timeseries(x)  # Should not raise

    x = np.array([[1, 2], [3, 4], [5, 6]])
    check_timeseries(x, expected_length=3, expected_dim=2)

    with pytest.raises(TypeError):
        check_timeseries([[1, 2], [3, 4]])

    x = np.array([1, 2, 3])
    with pytest.raises(ValueError):
        check_timeseries(x)

    x = np.array([[1, 2], [3, 4]])
    with pytest.raises(ValueError):
        check_timeseries(x, expected_length=5)

    x = np.array([[1, 2], [3, 4]])
    with pytest.raises(ValueError):
        check_timeseries(x, expected_dim=5)


def test_is_timeseries_valid():
    x = np.array([[1, 2], [3, 4]])
    assert is_timeseries(x) == True

    x = np.array([1, 2, 3])
    assert is_timeseries(x) == False


def test_check_multiseries():
    ts1 = np.array([[1, 2], [3, 4]])
    ts2 = np.array([[5, 6], [7, 8]])
    x = [ts1, ts2]
    check_multiseries(x)  # Should not raise

    ts1 = np.array([[1, 2], [3, 4]])
    ts2 = np.array([[5, 6], [7, 8]])
    x = [ts1, ts2]
    check_multiseries(x, expected_length=2, expected_dim=2)

    x = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
    check_multiseries(x)  # Should not raise

    x = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
    check_multiseries(x, expected_length=2, expected_dim=2)

    x = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
    with pytest.raises(ValueError):
        check_multiseries(x, expected_length=5)

    x = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
    with pytest.raises(ValueError):
        check_multiseries(x, expected_dim=5)

    x = np.array([[1, 2], [3, 4]])
    with pytest.raises(ValueError):
        check_multiseries(x)

    with pytest.raises(TypeError):
        check_multiseries(2504)

    ts1 = np.array([[1, 2], [3, 4]])
    ts2 = np.array([[5, 6], [7, 8]])
    x = [ts1, ts2]
    check_multiseries(x)

    # Test that it infers expected_dim from first series
    ts3 = np.array([[9, 10, 11], [12, 13, 14]])  # 3 features - should fail
    x_inconsistent = [ts1, ts3]
    with pytest.raises(ValueError):
        check_multiseries(x_inconsistent)

    x_invalid_type = (
        "According to all known laws of aviation, there is no way a bee should"
        "be able to fly. Its wings are too small to get its fat little body off"
        "the ground. The bee, of course, flies anyway. Because bees don't care"
        "what humans think is impossible. Yellow, black. Yellow, black."
    )
    with pytest.raises(TypeError):
        check_multiseries({x_invalid_type})


def test_is_multiseries():
    ts1 = np.array([[1, 2], [3, 4]])
    ts2 = np.array([[5, 6], [7, 8]])
    x = [ts1, ts2]
    assert is_multiseries(x) == True

    x = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
    assert is_multiseries(x) == True

    x = np.array([[1, 2], [3, 4]])
    assert is_multiseries(x) == False


def test_check_node_input():
    ts1 = np.array([[1, 2], [3, 4]])
    ts2 = np.array([[5, 6], [7, 8]])
    x = [ts1, ts2]
    check_node_input(x)

    x = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
    check_node_input(x)

    x = np.array([[1, 2], [3, 4]])
    check_node_input(x)

    x = np.array([1, 2, 3])
    with pytest.raises(ValueError):
        check_node_input(x)

    with pytest.raises(TypeError):
        check_node_input(12)

    ts1 = np.array([[1, 2], [3, 4]])  # 2 features
    ts2 = np.array([[5, 6], [7, 8]])  # 2 features
    x = [ts1, ts2]
    check_node_input(x)

    # Test dimension consistency enforcement
    ts3 = np.array([[9, 10, 11], [12, 13, 14]])  # 3 features
    x_inconsistent = [ts1, ts3]
    with pytest.raises(ValueError):
        check_node_input(x_inconsistent)


def test_is_node_input():
    x = np.array([[1, 2], [3, 4]])
    assert is_node_input(x) == True

    x = np.array([1, 2, 3])
    assert is_node_input(x) == False


def test_check_model_timestep():
    """Test check_model_timestep with mapping input."""
    x1 = np.array([1, 2])
    x2 = np.array([3, 4])
    x = {"node1": x1, "node2": x2}
    check_model_timestep(x)

    x1 = np.array([1, 2])
    x2 = np.array([3, 4])
    x = {"node1": x1, "node2": x2}
    check_model_timestep(x, expected_inputs=["node1", "node2"])

    x1 = np.array([1, 2])
    x = {"node1": x1}
    with pytest.raises(ValueError):
        check_model_timestep(x, expected_inputs=["node2"])

    x1 = np.array([1, 2])
    x2 = np.array([3, 4, 5])
    x = {"node1": x1, "node2": x2}
    expected_dim = {"node1": 2, "node2": 3}
    check_model_timestep(x, expected_dim=expected_dim)

    x1 = np.array([1, 2])
    x2 = np.array([3, 4])
    x = {"node1": x1, "node2": x2}
    check_model_timestep(x, expected_dim=2)

    x = np.array([1, 2, 3])
    check_model_timestep(x)

    x = np.array([1, 2])
    expected_dim = {"node1": 2}
    check_model_timestep(x, expected_dim=expected_dim)

    x = np.array([1, 2])
    check_model_timestep(x, expected_dim=2)

    with pytest.raises(TypeError):
        check_model_timestep(12)


def test_is_model_timestep():
    x = np.array([1, 2, 3])
    assert is_model_timestep(x) == True

    assert is_model_timestep(12) == False


def test_check_model_input():
    ts1 = np.array([[1, 2], [3, 4]])
    ts2 = np.array([[5, 6], [7, 8]])
    x = {"node1": ts1, "node2": ts2}
    check_model_input(x)

    ts1 = np.array([[1, 2], [3, 4]])
    ts2 = np.array([[5, 6, 7], [8, 9, 10]])
    x = {"node1": ts1, "node2": ts2}
    expected_dim = {"node1": 2, "node2": 3}
    check_model_input(x, expected_dim=expected_dim)

    x = np.array([[1, 2], [3, 4]])
    check_model_input(x)

    ts1 = np.array([[1, 2], [3, 4]])
    ts2 = np.array([[5, 6], [7, 8]])
    x = [ts1, ts2]
    check_model_input(x)

    x = np.array([[1, 2], [3, 4]])
    expected_dim = {"node1": 2}
    check_model_input(x, expected_dim=expected_dim)

    x = np.array([[1, 2], [3, 4]])
    expected_dim = {"node1": 2, "node2": 3}
    with pytest.raises(TypeError):
        check_model_input(x, expected_dim=expected_dim)

    with pytest.raises(TypeError):
        check_model_input(12)


def test_is_model_input():
    x = np.array([[1, 2], [3, 4]])
    assert is_model_input(x) == True

    assert is_model_input(12) == False



"""
==============================================================
Hyperparameter optimization helpers (:mod:`reservoirpy.hyper`)
==============================================================

Utility module for optimizing hyperparameters, with support for both
*hyperopt*-based search and CPU-parallelized custom search using *joblib*.

Note
----

This module is meant to be used alongside *hyperopt*
and *Matplotlib* packages, which are not installed
with ReservoirPy by default. Before using the
:py:mod:`reservoirpy.hyper` module, consider installing
these packages if they are not already installed.

The function `parallel_research` of this module
do not rely on *hyperopt* and can be used independently.


.. autosummary::
    :toctree: generated/

    research
    parallel_research
    plot_hyperopt_report
"""
from ._hyperplot import plot_hyperopt_report
from ._hypersearch import research
from ._parallel_hypersearch import parallel_research

__all__ = ["research", "parallel_research", "plot_hyperopt_report"]



"""
===========================================================
Metrics and observables (:py:mod:`reservoirpy.observables`)
===========================================================

Metrics and observables for Reservoir Computing:

.. autosummary::
   :toctree: generated/

    spectral_radius
    mse
    rmse
    nrmse
    rsquare
    memory_capacity
    effective_spectral_radius
"""

from copy import deepcopy

# Author: Nathan Trouvain at 01/06/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from typing import Literal, Optional, Union

import numpy as np
from scipy import linalg
from scipy.sparse import issparse
from scipy.sparse.linalg import eigs

from .type import Weights
from .utils.random import rand_generator


def _check_arrays(y_true, y_pred):
    y_true_array = np.asarray(y_true)
    y_pred_array = np.asarray(y_pred)

    if not y_true_array.shape == y_pred_array.shape:
        raise ValueError(
            f"Shape mismatch between y_true and y_pred: "
            f"{y_true_array.shape} != {y_pred_array.shape}"
        )

    return y_true_array, y_pred_array


def spectral_radius(W: Weights, maxiter: Optional[int] = None) -> float:
    """Compute the spectral radius of a matrix `W`.

    Spectral radius is defined as the maximum absolute
    eigenvalue of `W`.

    Parameters
    ----------
    W : array-like (sparse or dense) of shape (N, N)
        Matrix from which the spectral radius will
        be computed.

    maxiter : int, optional
        Maximum number of Arnoldi update iterations allowed.
        By default, is equal to `W.shape[0] * 20`.
        See `Scipy documentation <https://docs.scipy.org/
        doc/scipy/reference/generated/scipy.sparse.linalg.eigs.html>`_
        for more informations.

    Returns
    -------
    float
        Spectral radius of `W`.

    Raises
    ------
    ArpackNoConvergence
        When computing spectral radius on large
        sparse matrices, it is possible that the
        Fortran ARPACK algorithm used to compute
        eigenvalues don't converge towards precise
        values. To avoid this problem, set the `maxiter`
        parameter to an higher value. Be warned that
        this may drastically increase the computation
        time.

    Examples
    --------
    >>> from reservoirpy.observables import spectral_radius
    >>> from reservoirpy.mat_gen import normal
    >>> W = normal(1000, 1000, degree=8)
    >>> print(spectral_radius(W))
    2.8758915077733564
    """
    if issparse(W):
        if maxiter is None:
            maxiter = W.shape[0] * 20

        return max(
            abs(
                eigs(
                    W,
                    k=1,
                    which="LM",
                    maxiter=maxiter,
                    return_eigenvectors=False,
                    v0=np.ones(W.shape[0], W.dtype),
                )
            )
        )

    return max(abs(linalg.eig(W)[0]))


def mse(y_true: np.ndarray, y_pred: np.ndarray, dimensionwise: bool = False) -> float:
    """Mean squared error metric:

    .. math::

        \\frac{\\sum_{i=0}^{N-1} (y_i - \\hat{y}_i)^2}{N}

    Parameters
    ----------
    y_true : array-like of shape (N, features)
        Ground truth values.
    y_pred : array-like of shape (N, features)
        Predicted values.
    dimensionwise: boolean, optional
        If True, return a mean squared error for each dimension of the timeseries

    Returns
    -------
    float
        Mean squared error.
    If `dimensionwise` is True, returns a Numpy array of shape $(features, )$.

    Examples
    --------
    >>> from reservoirpy.nodes import ESN
    >>> from reservoirpy.datasets import mackey_glass, to_forecasting
    >>> x_train, x_test, y_train, y_test = to_forecasting(mackey_glass(1000), test_size=0.2)
    >>> y_pred = ESN(units=100, sr=1).fit(x_train, y_train).run(x_test)

    >>> from reservoirpy.observables import mse
    >>> print(mse(y_true=y_test, y_pred=y_pred))
    0.03962918253990291
    """
    y_true_array, y_pred_array = _check_arrays(y_true, y_pred)

    if dimensionwise:
        if len(y_true_array.shape) == 3:
            axis = (0, 1)
        else:
            axis = 0
    else:
        axis = None
    return np.mean((y_true_array - y_pred_array) ** 2, axis=axis)


def rmse(y_true: np.ndarray, y_pred: np.ndarray, dimensionwise: bool = False) -> float:
    """Root mean squared error metric:

    .. math::

        \\sqrt{\\frac{\\sum_{i=0}^{N-1} (y_i - \\hat{y}_i)^2}{N}}

    Parameters
    ----------
    y_true : array-like of shape (N, features)
        Ground truth values.
    y_pred : array-like of shape (N, features)
        Predicted values.
    dimensionwise: boolean, optional
        If True, return a mean squared error for each dimension of the timeseries

    Returns
    -------
    float
        Root mean squared error.
    If `dimensionwise` is True, returns a Numpy array of shape $(features, )$.

    Examples
    --------
    >>> from reservoirpy.nodes import Reservoir, Ridge
    >>> model = Reservoir(units=100, sr=1) >> Ridge(ridge=1e-8)

    >>> from reservoirpy.datasets import mackey_glass, to_forecasting
    >>> x_train, x_test, y_train, y_test = to_forecasting(mackey_glass(1000), test_size=0.2)
    >>> y_pred = model.fit(x_train, y_train).run(x_test)

    >>> from reservoirpy.observables import rmse
    >>> print(rmse(y_true=y_test, y_pred=y_pred))
    0.00034475744480521534
    """
    return np.sqrt(mse(y_true, y_pred, dimensionwise=dimensionwise))


def nrmse(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    norm: Literal["minmax", "var", "mean", "q1q3"] = "minmax",
    norm_value: Optional[float] = None,
    dimensionwise: bool = False,
) -> float:
    """Normalized root mean squared error metric:

    .. math::

       \\frac{1}{\\lambda} * \\sqrt{\\frac{\\sum_{i=0}^{N-1} (y_i - \\hat{y}_i)^2}{N}}

    where :math:`\\lambda` may be:
        - :math:`\\max y - \\min y` (Peak-to-peak amplitude) if ``norm="minmax"``;
        - :math:`\\mathrm{Var}(y)` (variance over time) if ``norm="var"``;
        - :math:`\\mathbb{E}[y]` (mean over time) if ``norm="mean"``;
        - :math:`Q_{3}(y) - Q_{1}(y)` (quartiles) if ``norm="q1q3"``;
        - or any value passed to ``norm_value``.

    Parameters
    ----------
    y_true : array-like of shape (N, features)
        Ground truth values.
    y_pred : array-like of shape (N, features)
        Predicted values.
    norm : {"minmax", "var", "mean", "q1q3"}, default to "minmax"
        Normalization method.
    norm_value : float, optional
        A normalization factor. If set, will override the ``norm`` parameter.
    dimensionwise: boolean, optional
        If True, return a mean squared error for each dimension of the timeseries

    Returns
    -------
    float
        Normalized root mean squared error.
    If `dimensionwise` is True, returns a Numpy array of shape $(features, )$.

    Examples
    --------
    >>> from reservoirpy.nodes import Reservoir, Ridge
    >>> model = Reservoir(units=100, sr=1) >> Ridge(ridge=1e-8)

    >>> from reservoirpy.datasets import mackey_glass, to_forecasting
    >>> x_train, x_test, y_train, y_test = to_forecasting(mackey_glass(1000), test_size=0.2)
    >>> y_pred = model.fit(x_train, y_train).run(x_test)

    >>> from reservoirpy.observables import nrmse
    >>> print(nrmse(y_true=y_test, y_pred=y_pred, norm="var"))
    0.007854318015438394
    """
    error = rmse(y_true, y_pred, dimensionwise=dimensionwise)
    if norm_value is not None:
        return error / norm_value

    else:
        y_true_array, y_pred_array = _check_arrays(y_true, y_pred)

        if dimensionwise:
            if len(y_true_array.shape) == 3:
                axis = (0, 1)
            else:
                axis = 0
        else:
            axis = None

        norms = {
            "minmax": lambda y: np.ptp(y, axis=axis),
            "var": lambda y: y.var(axis=axis),
            "mean": lambda y: y.mean(axis=axis),
            "q1q3": lambda y: np.quantile(y, 0.75, axis=axis)
            - np.quantile(y, 0.25, axis=axis),
        }

        if norms.get(norm) is None:
            raise ValueError(
                f"Unknown normalization method. "
                f"Available methods are {list(norms.keys())}."
            )
        else:
            return error / norms[norm](y_true_array)


def rsquare(
    y_true: np.ndarray, y_pred: np.ndarray, dimensionwise: bool = False
) -> float:
    """Coefficient of determination :math:`R^2`:

    .. math::

        1 - \\frac{\\sum^{N-1}_{i=0} (y - \\hat{y})^2}
        {\\sum^{N-1}_{i=0} (y - \\bar{y})^2}

    where :math:`\\bar{y}` is the mean value of ground truth.

    Parameters
    ----------
    y_true : array-like of shape (N, features)
        Ground truth values.
    y_pred : array-like of shape (N, features)
        Predicted values.
    dimensionwise: boolean, optional
        If True, return a mean squared error for each dimension of the timeseries

    Returns
    -------
    float
        Coefficient of determination.
    If `dimensionwise` is True, returns a Numpy array of shape $(features, )$.

    Examples
    --------
    >>> from reservoirpy.nodes import Reservoir, Ridge
    >>> model = Reservoir(units=100, sr=1) >> Ridge(ridge=1e-8)

    >>> from reservoirpy.datasets import mackey_glass, to_forecasting
    >>> x_train, x_test, y_train, y_test = to_forecasting(mackey_glass(1000), test_size=0.2)
    >>> y_pred = model.fit(x_train, y_train).run(x_test)

    >>> from reservoirpy.observables import rsquare
    >>> print(rsquare(y_true=y_test, y_pred=y_pred))
    0.9999972921653904
    """
    y_true_array, y_pred_array = _check_arrays(y_true, y_pred)

    if dimensionwise:
        if len(y_true_array.shape) == 3:
            axis = (0, 1)
        else:
            axis = 0
    else:
        axis = None

    d = (y_true_array - y_pred_array) ** 2
    D = (y_true_array - y_true_array.mean(axis=axis)) ** 2
    return 1 - np.sum(d, axis=axis) / np.sum(D, axis=axis)


def memory_capacity(
    model: "Model",
    k_max: int,
    as_list: bool = False,
    series: Optional[np.ndarray] = None,
    test_size: Union[int, float] = 0.2,
    seed: Optional[Union[int, np.random.RandomState, np.random.Generator]] = None,
):
    """Memory Capacity of a model

    The Memory Capacity [1]_ (MC) measure is defined as:

    .. math::
        MC = \\sum_{k=1}^{k_{max}} MC_k

    where:

    .. math::
        MC_k = \\rho^2(u(t-k), y_k(t)) = {cov^2[u(t-k), y_k(t)] \\over var(u(t-k)) \\cdot var(y_k(t))}

    This measure is commonly used in reservoir computing to evaluate the ability of a network to
    recall the previous timesteps. By default, the timeseries :math:`u` is an i.i.d. uniform signal in [-0.8, 0.8].

    Parameters
    ----------
    model : :class:`reservoirpy.Model`
        A ReservoirPy model on which the memory capacity is tested.
        Must have only one input and output node.
    k_max : int
        Maximum time lag between input and output.
        A common rule of thumb is to choose `k_max = 2*reservoir.units`.
    as_list: bool, optional, defaults to False
        If True, returns an array in which `out[k]` :math:`= MC_{k+1}`
    series: array of shape (timesteps, 1), optional
        If specified, is used as the timeseries :math:`u`.
    test_size : int or float
        Number of timesteps for the training phase. Can also be specified
        as a float ratio.
    seed : int or :py:class:`numpy.random.Generator`, optional
        Random state seed for reproducibility.

    Returns
    -------
    float, between 0 and `k_max`.
    If `as_list` is set to True, returns an array of shape `(k_max, )`.

    Examples
    --------
    >>> from reservoirpy.nodes import Reservoir, Ridge
    >>> from reservoirpy.observables import memory_capacity
    >>> model = Reservoir(100, sr=1, seed=1) >> Ridge(ridge=1e-4)
    >>> mcs = memory_capacity(model, k_max=50, as_list=True, seed=1)
    >>> print(f"Memory capacity of the model: {np.sum(mcs):.4}")
    Memory capacity of Model-0: 12.77

    .. plot::

        from reservoirpy.nodes import Reservoir, Ridge
        from reservoirpy.observables import memory_capacity
        model = Reservoir(100, sr=1, seed=1) >> Ridge(ridge=1e-4)
        mcs = memory_capacity(model, k_max=50, as_list=True, seed=1)
        plt.figure(figsize=(8, 4))
        plt.plot(np.arange(1, 51), mcs)
        plt.grid(); plt.xlabel("Time lag (timestep)"); plt.ylabel("$MC_k$")
        plt.show()

    References
    ----------
    .. [1] Jaeger, H. (2001). Short term memory in echo state networks.
    """
    from numpy.lib.stride_tricks import sliding_window_view

    # Task definition
    if series is None:
        rng = rand_generator(seed)
        series = rng.uniform(low=-0.8, high=0.8, size=(10 * k_max, 1))

    if isinstance(test_size, float) and test_size < 1 and test_size >= 0:
        test_len = round(series.shape[0] * test_size)
    elif isinstance(test_size, int):
        test_len = test_size
    else:
        raise ValueError(
            "invalid test_size argument: "
            "test_size can be an integer or a float "
            f"in [0, 1[, but is {test_size}."
        )

    # sliding_window_view creates a matrix of the same
    # timeseries with an incremental shift on each column
    dataset = sliding_window_view(series[:, 0], k_max + 1)[:, ::-1]
    X_train = dataset[:-test_len, :1]
    X_test = dataset[-test_len:, :1]
    Y_train = dataset[:-test_len, 1:]
    Y_test = dataset[-test_len:, 1:]
    # Model
    model_clone = deepcopy(model)
    model_clone.fit(X_train, Y_train, warmup=k_max)
    Y_pred = model_clone.run(X_test)

    # u[t-k] - z_k[t] square correlation
    capacities = np.square(
        [
            np.corrcoef(y_pred, y_test, rowvar=False)[1, 0]
            for y_pred, y_test in zip(Y_pred.T, Y_test.T)
        ]
    )

    if as_list:
        return capacities
    else:
        return np.sum(capacities)


def effective_spectral_radius(
    W: np.ndarray, lr: float = 1.0, maxiter: Optional[int] = None
):
    """Effective spectral radius

    The effective spectral radius is defined as the maximal singular value of the matrix
    :math:`lr \cdot W + (1-lr) \cdot I_{n}`.

    This concept was first introduced by Jaeger & al. [1]_, with an important result on leaky echo
    state networks:

    Supposing:

    #. The activation function is `tanh`
    #. There is no added noise inside the reservoir (`noise_rc = noise_in = 0.0`)
    #. There is no feedback
    #. There is no bias inside the reservoir (`bias = 0`)

    Then, if the effective spectral radius exceeds 1, the ESN does not have the echo state property.

    Parameters
    ----------
    W : array of shape `(units, units)`
        Adjacency matrix of a reservoir
    lr : float
        Leak rate
    maxiter : int, optional
        Maximum number of Arnoldi update iterations allowed.
        By default, is equal to `W.shape[0] * 20`.
        See `Scipy documentation <https://docs.scipy.org/
        doc/scipy/reference/generated/scipy.sparse.linalg.eigs.html>`_
        for more informations.

    Returns
    -------
    float
        Spectral radius of :math:`lr \cdot W + (1-lr) \cdot I_{n}`.

    Raises
    ------
    ArpackNoConvergence
        When computing spectral radius on large
        sparse matrices, it is possible that the
        Fortran ARPACK algorithm used to compute
        eigenvalues don't converge towards precise
        values. To avoid this problem, set the `maxiter`
        parameter to an higher value. Be warned that
        this may drastically increase the computation
        time.

    Examples
    --------

    >>> from reservoirpy.observables import spectral_radius, effective_spectral_radius
    >>> from reservoirpy.mat_gen import uniform
    >>> W = uniform(100, 100, sr=0.5, seed=0)
    >>> lr = 0.5
    >>> print(f"{spectral_radius(W)=:.3}")
    >>> print(f"{effective_spectral_radius(W, lr=lr)=:.3}")
    spectral_radius(W)=0.5
    effective_spectral_radius(W, lr=lr)=0.701

    References
    ----------
    .. [1] Jaeger, H., Lukoševičius, M., Popovici, D., & Siewert, U. (2007).
        Optimization and applications of echo state networks with leaky-integrator
        neurons. Neural networks, 20(3), 335-352.
    """
    units = W.shape[0]
    return spectral_radius(W=lr * W + (1 - lr) * np.eye(units), maxiter=maxiter)



from typing import Callable, Optional, Tuple, Union

import numpy as np

from reservoirpy.utils.data_validation import check_node_input

from ..mat_gen import zeros
from ..node import OnlineNode
from ..type import NodeInput, State, Timeseries, Timestep, Weights, is_array


class RLS(OnlineNode):
    """Single layer of neurons learning connections using Recursive Least Squares
    algorithm.

    The learning rules is well described in [1]_.
    The forgetting factor version of the RLS algorithm used here is described in [2]_.

    Parameters
    ----------
    alpha : float or Python generator or iterable, default to 1e-6
        Diagonal value of matrix P.
    Wout : callable or array-like of shape (units, targets), default to :py:func:`~reservoirpy.mat_gen.zeros`
        Output weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    bias : callable or array-like of shape (units, 1), default to :py:func:`~reservoirpy.mat_gen.zeros`
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    fit_bias : bool, default to True
        If True, then a bias parameter will be learned along with output weights.
    forgetting : float, default to 1.0
        The forgetting factor controls the weight given to past observations in the RLS update.
        A value less than 1.0 gives more weight to recent observations.
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    output_dim : int, optional
        Number of units in the readout, can be inferred at first call.
    name : str, optional
        Node name.

    References
    ----------

    .. [1] Sussillo, D., & Abbott, L. F. (2009). Generating Coherent Patterns of
           Activity from Chaotic Neural Networks. Neuron, 63(4), 544–557.
           https://doi.org/10.1016/j.neuron.2009.07.018

    .. [2] Waegeman, T., Wyffels, F., & Schrauwen, B. (2012). Feedback Control by Online
           Learning an Inverse Model. IEEE Transactions on Neural Networks and Learning
           Systems, 23(10), 1637–1648. https://doi.org/10.1109/TNNLS.2012.2208655

    Examples
    --------
    >>> x = np.random.normal(size=(100, 3))
    >>> noise = np.random.normal(scale=0.1, size=(100, 1))
    >>> y = x @ np.array([[10], [-0.2], [7.]]) + noise + 12.

    >>> from reservoirpy.nodes import RLS
    >>> rls_node = RLS(alpha=1e-1)

    >>> _ = rls_node.partial_fit(x[:5], y[:5])
    >>> print(rls_node.Wout.T, rls_node.bias)
    [[ 9.90731641 -0.06884784  6.87944632]] [[12.07802068]]
    >>> _ = rls_node.partial_fit(x[5:], y[5:])
    >>> print(rls_node.Wout.T, rls_node.bias)
    [[ 9.99223366 -0.20499636  6.98924066]] [[12.01128622]]
    """

    #: Learned output weights (:math:`\\mathbf{W}_{out}`).
    Wout: Weights
    #: Learned bias (:math:`\\mathbf{b}`).
    bias: Weights
    #: Matrix :math:`\\mathbf{P}` of RLS rule.
    P: Weights
    #: Diagonal value of matrix P (:math:`\\alpha`) (:math:`1\\cdot 10^{-6}` by default).
    alpha: float
    #: If True, learn a bias term (True by default).
    fit_bias: bool
    #: Forgetting factor (:math:`\\lambda`) (:math:`1` by default).
    forgetting: float

    def __init__(
        self,
        alpha: float = 1e-6,
        Wout: Union[Weights, Callable] = zeros,
        bias: Union[Weights, Callable] = zeros,
        fit_bias: bool = True,
        forgetting: float = 1.0,
        input_dim: Optional[int] = None,
        output_dim: Optional[int] = None,
        name: Optional[str] = None,
    ):
        self.alpha = alpha
        self.Wout = Wout
        self.bias = bias
        self.fit_bias = fit_bias
        self.forgetting = forgetting
        self.name = name

        self.initialized = False
        self.state = {}

        # set input_dim/output_dim (if possible)
        self.input_dim = input_dim
        self.output_dim = output_dim
        if is_array(Wout):
            if input_dim is not None and Wout.shape[0] != input_dim:
                raise ValueError(
                    f"Both 'input_dim' and 'Wout' are set but their dimensions doesn't "
                    f"match: {input_dim} != {Wout.shape[0]}."
                )
            self.input_dim = Wout.shape[0]
            if output_dim is not None and Wout.shape[1] != output_dim:
                raise ValueError(
                    f"Both 'output_dim' and 'Wout' are set but their dimensions doesn't "
                    f"match: {output_dim} != {Wout.shape[1]}."
                )
            self.output_dim = Wout.shape[1]
        if is_array(bias):
            if output_dim is not None and bias.shape[0] != output_dim:
                raise ValueError(
                    f"Both 'output_dim' and 'Wout' are set but their dimensions doesn't "
                    f"match: {output_dim} != {bias.shape[0]}."
                )
            self.output_dim = bias.shape[0]

    def _run(self, state: State, x: Timeseries) -> Tuple[State, Timeseries]:
        out = x @ self.Wout + self.bias  # (len, in) @ (in, out) + (out,)
        return {"out": out[-1]}, out

    def _step(self, state: State, x: Timestep) -> State:
        return {"out": x @ self.Wout + self.bias}  # (in, ) @ (in, out) + (out,)

    def initialize(
        self,
        x: Optional[Union[NodeInput, Timestep]],
        y: Optional[Union[NodeInput, Timestep]] = None,
    ):
        self._set_input_dim(x)
        self._set_output_dim(y)

        # initialize matrices
        if isinstance(self.Wout, Callable):
            self.Wout = self.Wout(self.input_dim, self.output_dim)
        if isinstance(self.bias, Callable):
            self.bias = self.bias(self.output_dim)
        self.P = np.eye(self.input_dim) / self.alpha
        self.S = 0

        self.initialized = True

    def _learning_step(
        self,
        x: Timestep,
        y: Timestep,
    ):
        """
        Wout: np.ndarray (in, out)
        bias: np.ndarray (out,)
        P: np.ndarray (in, in)
        x: np.ndarray (in,)
        y: np.ndarray (out,)
        Returns
        (Wout_next, bias_next, P_next, S_next), y_pred
        """
        Wout: Weights = self.Wout
        bias: Weights = self.bias
        P: np.ndarray = self.P
        forgetting: float = self.forgetting
        S: float = self.S

        Px = P @ x  # (in,)
        dP = -np.outer(Px, Px) / (forgetting + x @ Px)  # (in, in)
        P_next = (P + dP) / forgetting
        S_next = forgetting * S + 1

        prediction = x @ Wout + bias  # (out,) = (in,) @ (in, out) + (out,)
        error = prediction - y  # (out,)
        dWout = -np.outer(P_next @ x, error)  # (in, out)
        Wout_next = Wout + dWout  # (in, out)

        if self.fit_bias:
            bias_next = (S_next * bias - error) / S_next
        else:
            bias_next = bias
        y_pred = x @ Wout_next + bias

        self.Wout = Wout_next
        self.bias = bias_next
        self.P = P_next
        self.S = S_next
        return y_pred

    def partial_fit(self, x: Timeseries, y: Timeseries):
        check_node_input(x, expected_dim=self.input_dim)
        check_node_input(y)

        if not self.initialized:
            self.initialize(x, y)

        n_timesteps = x.shape[-2]
        out_dim = y.shape[-1]
        y_pred = np.empty((n_timesteps, out_dim))
        for i, (x_, y_) in enumerate(zip(x, y)):
            y_pred_ = self._learning_step(x_, y_)
            y_pred[i] = y_pred_

        self.state = {"out": y_pred_}
        return y_pred



# Author: Nathan Trouvain at 24/02/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest
from numpy.testing import assert_allclose

from ..intrinsic_plasticity import IPReservoir


def test_ip_init():
    res = IPReservoir(100, input_dim=5)
    x = np.ones((10, 5))

    res.initialize(x)

    assert res.W.shape == (100, 100)
    assert res.Win.shape == (100, 5)
    assert_allclose(res.a, np.ones((100,)))
    assert_allclose(res.b, np.zeros((100,)))

    res = IPReservoir(100)

    out = res.run(x)

    assert out.shape == (10, 100)
    assert res.W.shape == (100, 100)
    assert res.Win.shape == (100, 5)
    assert_allclose(res.a, np.ones((100,)))
    assert_allclose(res.b, np.zeros((100,)))

    with pytest.raises(ValueError):
        res = IPReservoir(100, activation="identity")


def test_intrinsic_plasticity():

    x = np.random.normal(size=(100, 5))
    X = [x[:10], x[:20]]

    res = IPReservoir(100, activation="tanh", epochs=2)

    res.fit(x)
    res.fit(X)

    assert res.a.shape == (100,)
    assert res.b.shape == (100,)

    res = IPReservoir(100, activation="sigmoid", epochs=1, mu=0.1)

    res.fit(x)
    res.fit(X)

    assert res.a.shape == (100,)
    assert res.b.shape == (100,)

    res.fit(x)
    res.fit(X)

    assert res.a.shape == (100,)
    assert res.b.shape == (100,)


def test_ip_model():
    x = np.random.normal(size=(100, 5))
    y = np.random.normal(size=(100, 2))
    X = [x[:10], x[:20]]
    Y = [y[:10], y[:20]]

    res = IPReservoir(100, activation="tanh", epochs=2, seed=1234)

    res.fit(X, Y)

    res2 = IPReservoir(100, activation="tanh", epochs=2, seed=1234)
    res2.fit(X)

    assert_allclose(res.a, res2.a)
    assert_allclose(res.b, res2.b)



# Author: Paul BERNARD at 01/01/2024 <paul.bernard@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

import numpy as np
import pytest
from sklearn.decomposition import PCA
from sklearn.gaussian_process.kernels import DotProduct
from sklearn.linear_model import (
    ElasticNet,
    Lars,
    Lasso,
    LassoCV,
    LassoLars,
    LinearRegression,
    LogisticRegression,
    MultiTaskLassoCV,
    OrthogonalMatchingPursuitCV,
    PassiveAggressiveClassifier,
    Perceptron,
    Ridge,
    RidgeClassifier,
    SGDClassifier,
    SGDRegressor,
)
from sklearn.neural_network import MLPClassifier, MLPRegressor

import reservoirpy
from reservoirpy.nodes import ScikitLearnNode


def test_fail_non_predictors():
    with pytest.raises(AttributeError):
        _ = ScikitLearnNode(PCA)
    with pytest.raises(AttributeError):
        _ = ScikitLearnNode(DotProduct)


def test_scikitlearn_initializer():

    with pytest.raises(TypeError):
        _ = ScikitLearnNode(LinearRegression).initialize()

    with pytest.raises(TypeError):
        _ = ScikitLearnNode(LinearRegression).initialize(np.ones((100, 2)))

    _ = ScikitLearnNode(LinearRegression).initialize(np.ones((100, 2)), np.ones((100, 2)))

    linear_regressor = ScikitLearnNode(LinearRegression, output_dim=2, positive=False)
    linear_regressor.initialize(np.ones((100, 2)))

    assert linear_regressor.model_kwargs == {"positive": False}


# Note that a different seed may fail the tests
@pytest.mark.parametrize(
    "model, model_hypers",
    [
        (LogisticRegression, {"random_state": 2341}),
        (PassiveAggressiveClassifier, {"random_state": 2341}),
        (Perceptron, {"random_state": 2341}),
        (RidgeClassifier, {"random_state": 2341}),
        (SGDClassifier, {"random_state": 2341}),
        (MLPClassifier, {"tol": 1e-3, "random_state": 2341}),
    ],
)
def test_scikitlearn_classifiers(model, model_hypers):
    rng = np.random.default_rng(seed=2341)

    X_train = rng.normal(0, 1, size=(100, 2))
    y_train = (X_train[:, 0:1] > 0.0).astype(np.float16)
    X_test = rng.normal(0, 1, size=(10, 2))
    y_test = (X_test[:, 0:1] > 0.0).astype(np.float16)

    scikit_learn_node = ScikitLearnNode(model=model, **model_hypers)

    scikit_learn_node.fit(X_train, y_train)
    y_pred = scikit_learn_node.run(X_test)
    assert y_pred.shape == y_test.shape
    assert np.all(y_pred == y_test)


@pytest.mark.parametrize(
    "model, model_hypers",
    [
        (LinearRegression, {}),
        (Ridge, {"random_state": 2341}),
        (SGDRegressor, {"random_state": 2341}),
        (ElasticNet, {"alpha": 1e-4, "random_state": 2341}),
        (Lars, {"random_state": 2341}),
        (Lasso, {"alpha": 1e-4, "random_state": 2341}),
        (LassoLars, {"alpha": 1e-4, "random_state": 2341}),
        (OrthogonalMatchingPursuitCV, {}),
        (MLPRegressor, {"tol": 1e-4, "random_state": 2341}),
    ],
)
def test_scikitlearn_regressors_monooutput(model, model_hypers):
    rng = np.random.default_rng(seed=2341)
    X_train = list(rng.normal(0, 1, size=(30, 100, 2)))
    y_train = [(x[:, 0:1] + x[:, 1:2]).astype(np.float16) for x in X_train]
    X_test = rng.normal(0, 1, size=(100, 2))
    y_test = (X_test[:, 0:1] + X_test[:, 1:2]).astype(np.float16)

    scikit_learn_node = ScikitLearnNode(model=model, **model_hypers)

    scikit_learn_node.fit(X_train, y_train)
    y_pred = scikit_learn_node.run(X_test)
    assert y_pred.shape == y_test.shape
    mse = np.mean(np.square(y_pred - y_test))
    assert mse < 2e-3

    timestep = scikit_learn_node(X_test[0])
    assert timestep.ndim == 1


def test_scikitlearn_multioutput():
    rng = np.random.default_rng(seed=2341)
    X_train = rng.normal(0, 1, size=(200, 3))
    y_train = X_train @ np.array([[0, 1, 0], [0, 1, 1], [-1, 0, 1]])
    X_test = rng.normal(0, 1, size=(100, 3))

    lasso = ScikitLearnNode(model=LassoCV, random_state=2341).fit(X_train, y_train).fit(X_train, y_train)  # fit twice
    lasso_pred = lasso.run(X_test)

    mt_lasso = ScikitLearnNode(model=MultiTaskLassoCV, random_state=2341).fit(X_train, y_train)
    mt_lasso_pred = mt_lasso.run(X_test)

    assert type(lasso.instances) is list
    assert type(mt_lasso.instances) is not list

    coef_single = [
        lasso.instances[0].coef_,
        lasso.instances[1].coef_,
        lasso.instances[2].coef_,
    ]
    coef_multitask = mt_lasso.instances.coef_
    assert np.linalg.norm(coef_single[0] - coef_multitask[0]) < 1e-3
    assert np.linalg.norm(coef_single[1] - coef_multitask[1]) < 1e-3
    assert np.linalg.norm(coef_single[2] - coef_multitask[2]) < 1e-3

    assert lasso_pred.shape == mt_lasso_pred.shape == (100, 3)
    assert np.linalg.norm(mt_lasso_pred - lasso_pred) < 1e-2


def test_scikitlearn_multiseries():
    rng = np.random.default_rng(seed=2341)
    X_train = rng.normal(0, 1, size=(5, 50, 3))
    y_train = X_train @ np.array([[0, 1, 0], [0, 1, 1], [-1, 0, 1]])
    X_test = rng.normal(0, 1, size=(2, 100, 3))

    lasso = ScikitLearnNode(model=LassoCV, random_state=2341)
    lasso.fit(X_train, y_train)
    lasso_pred = lasso.run(X_test)

    assert lasso_pred.shape == (2, 100, 3)

    rng = np.random.default_rng(seed=2341)
    X_train = list(rng.normal(0, 1, size=(5, 200, 3)))
    y_train = X_train @ np.array([[0, 1, 0], [0, 1, 1], [-1, 0, 1]])
    X_test = list(rng.normal(0, 1, size=(2, 100, 3)))

    lasso = ScikitLearnNode(model=LassoCV, random_state=2341)
    lasso.fit(X_train, y_train)
    lasso_pred = lasso.run(X_test)

    assert isinstance(lasso_pred, list)
    assert len(lasso_pred) == 2
    assert lasso_pred[0].shape == (100, 3)


def test_scikitlearn_reproductibility_random_state():
    rng = np.random.default_rng(seed=2341)
    X_train = rng.normal(0, 1, size=(100, 3))
    y_train = (X_train @ np.array([0.5, 1, 2])).reshape(-1, 1)
    X_test = rng.normal(0, 1, size=(100, 3))

    # Different scikit-learn random_states
    reservoirpy.set_seed(0)
    y_pred1 = ScikitLearnNode(model=SGDRegressor, random_state=1).fit(X_train, y_train).run(X_test)

    reservoirpy.set_seed(0)
    y_pred2 = ScikitLearnNode(model=SGDRegressor, random_state=2).fit(X_train, y_train).run(X_test)

    assert not np.all(y_pred1 == y_pred2)

    # Same scikit-learn random_states
    reservoirpy.set_seed(0)
    y_pred1 = ScikitLearnNode(model=SGDRegressor, random_state=1).fit(X_train, y_train).run(X_test)

    reservoirpy.set_seed(0)
    y_pred2 = ScikitLearnNode(model=SGDRegressor, random_state=1).fit(X_train, y_train).run(X_test)

    assert np.all(y_pred1 == y_pred2)

    # Same scikit-learn random_states (call)
    reservoirpy.set_seed(0)
    y_pred1 = ScikitLearnNode(model=SGDRegressor, random_state=1).fit(X_train, y_train)(X_test[0])

    reservoirpy.set_seed(0)
    y_pred2 = ScikitLearnNode(model=SGDRegressor, random_state=1).fit(X_train, y_train)(X_test[0])

    assert np.all(y_pred1 == y_pred2)


def test_scikitlearn_reproductibility_rpy_seed():
    rng = np.random.default_rng(seed=2341)
    X_train = rng.normal(0, 1, size=(100, 3))
    y_train = (X_train @ np.array([0.5, 1, 2])).reshape(-1, 1)
    X_test = rng.normal(0, 1, size=(100, 3))

    # Different ReservoirPy random generator
    reservoirpy.set_seed(1)
    y_pred1 = ScikitLearnNode(model=SGDRegressor).fit(X_train, y_train).run(X_test)

    reservoirpy.set_seed(2)
    y_pred2 = ScikitLearnNode(model=SGDRegressor).fit(X_train, y_train).run(X_test)

    assert not np.all(y_pred1 == y_pred2)

    # Same ReservoirPy random generator
    reservoirpy.set_seed(0)
    y_pred1 = ScikitLearnNode(model=SGDRegressor).fit(X_train, y_train).run(X_test)

    reservoirpy.set_seed(0)
    y_pred2 = ScikitLearnNode(model=SGDRegressor).fit(X_train, y_train).run(X_test)

    assert np.all(y_pred1 == y_pred2)



# Author: Nathan Trouvain at 08/07/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import pickle

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

from .dummy_nodes import AccumulateNode, Offline, OnlineUnsupervised, PlusNode


def test_pickling():
    plus_node = PlusNode(name="MyNode", h=12)
    pickled_node = pickle.dumps(plus_node)
    unpickled_node = pickle.loads(pickled_node)

    assert unpickled_node.name == "MyNode"
    assert plus_node.h == 12
    assert plus_node != unpickled_node


def test_node_init():
    plus_node = PlusNode()
    data = np.zeros((5,))

    res = plus_node(data)

    assert_array_equal(res, data + 1)
    assert plus_node.initialized
    assert plus_node.input_dim == 5
    assert plus_node.output_dim == 5
    assert plus_node.h == 1

    data = np.zeros((1, 8))

    with pytest.raises(ValueError):
        plus_node(data)


def test_node_call():
    plus_node = PlusNode(h=2)
    data = np.zeros((5,))
    res = plus_node(data)

    assert_array_equal(res, data + 2)
    assert plus_node.state is not None
    assert_array_equal(data + 2, plus_node.state["out"])


def test_node_dimensions():
    plus_node = PlusNode()
    data = np.zeros((5,))
    res = plus_node(data)

    # input size mismatch
    with pytest.raises(ValueError):
        data = np.zeros((6,))
        plus_node(data)

    # input size mismatch in run,
    # no matter how many timesteps are given
    with pytest.raises(ValueError):
        data = np.zeros((5, 6))
        plus_node.run(data)

    with pytest.raises(ValueError):
        data = np.zeros((1, 6))
        plus_node.run(data)

    # no timespans in call, only single timesteps
    with pytest.raises(ValueError):
        data = np.zeros((2, 5))
        plus_node(data)


def test_node_run():
    plus_node = AccumulateNode(h=2)
    data = np.zeros((3, 5))
    res = plus_node.run(data)
    expected = np.array([[2] * 5, [4] * 5, [6] * 5])

    assert_array_equal(res, expected)
    assert_array_equal(res[-1], plus_node.state["out"])

    # res2 = plus_node.run(data, stateful=False)
    # expected2 = np.array([[8] * 5, [10] * 5, [12] * 5])

    # assert_array_equal(res2, expected2)
    # assert_array_equal(res[-1][np.newaxis, :], plus_node.state["out"])

    # res3 = plus_node.run(data, reset=True)

    # assert_array_equal(res3, expected)
    # assert_array_equal(res[-1][np.newaxis, :], plus_node.state["out"])


def test_offline_fit():
    offline_node = Offline()
    X = np.ones((10, 5)) * 0.5
    Y = np.ones((10, 5))

    assert offline_node.b == 0

    offline_node.fit(X, Y)

    assert_allclose(offline_node.b, 50.0)

    Y = -np.ones((10, 5))

    offline_node.fit(X, Y)

    assert_allclose(offline_node.b, -50)

    X = [np.ones((10, 5)) * 2.0] * 3
    Y = [0.3 * np.ones((10, 5))] * 3

    offline_node.fit(X, Y)

    assert_allclose(offline_node.b, 0.3 * 10 * 5 * 3)

    offline_node.fit(np.array(X), np.array(Y))

    assert_allclose(offline_node.b, 0.3 * 10 * 5 * 3)


def test_unsupervised_fit():
    unsupervised_node = OnlineUnsupervised()
    X = np.ones((10, 5))

    assert unsupervised_node.b == 0

    unsupervised_node.fit(X)

    assert_allclose(unsupervised_node.b, 50.0)

    X = -np.ones((10, 5))

    unsupervised_node.fit(X)

    assert_allclose(unsupervised_node.b, -50)

    X = [0.3 * np.ones((10, 5))] * 3

    unsupervised_node.fit(X)

    assert_allclose(unsupervised_node.b, 0.3 * 10 * 5 * 3)

    unsupervised_node.fit(np.array(X))

    assert_allclose(unsupervised_node.b, 0.3 * 10 * 5 * 3)


def test_partial_fit_unsupervised():
    online_node = OnlineUnsupervised()
    X = np.ones((10, 5))

    assert online_node.b == 0

    online_node.partial_fit(X)

    assert_allclose(online_node.b, 50.0)

    X = np.ones((10, 5)) * 2.0

    online_node.partial_fit(X)

    assert_allclose(online_node.b, 150.0)

    X = [np.ones((10, 5)) * 2.0] * 3


# def test_train_raise():
#     online_node = OnlineUnsupervised()
#     X = [np.ones((10, 5)) * 2.0] * 3
#     Y = [np.ones((10, 5)) * 2.0] * 3

#     with pytest.raises(TypeError):
#         online_node.partial_fit(X, Y)


def test_node_bad_learning_method():
    online_node = OnlineUnsupervised()
    plus_node = PlusNode()
    offline_node = Offline()

    X = np.ones((10, 5))
    Y = np.ones((10, 5))

    with pytest.raises(AttributeError):
        plus_node.fit(X, Y)

    with pytest.raises(AttributeError):
        offline_node.partial_fit(X, Y)

    with pytest.raises(AttributeError):
        plus_node.partial_fit(X, Y)


# def test_offline_node_bad_warmup():
#     offline_node = Offline()
#     X = np.ones((10, 5))
#     Y = np.ones((10, 5))

#     with pytest.raises(ValueError):
#         offline_node.fit(X, Y, warmup=10)



from typing import Callable, Optional, Sequence, Union

import numpy as np

from ..activationsfunc import get_function, tanh
from ..mat_gen import bernoulli, normal, orthogonal
from ..node import Node
from ..type import NodeInput, State, Timestep, Weights, is_array
from ..utils import random


class ES2N(Node):
    """Edge of Stability Echo State Network.

    As first described in [1]_.

    The Edge of Stability Echo State Network (:math:`ES^2N`) model is similar to the reservoir
    equation, but adds an orthogonal transformation to the recurrent part of the equation.

    Reservoir neurons states, gathered in a vector :math:`\\mathbf{x}`, follow
    the update rule below:

    .. math::

        \\mathbf{x}[t+1] = (1 - \\mathrm{\\beta}) \\mathbf{O} * \\mathbf{x}[t] + \\mathrm{\\beta}
         * f(\\mathbf{W}_{in} \\cdot \\mathbf{u}[t+1]
          + \\mathbf{W} \\cdot \\mathbf{x}[t])

    where:
        - :math:`\\mathbf{x}` is the output activation vector of the reservoir;
        - :math:`\\mathbf{u}` is the input timeseries;
        - :math:`f` and :math:`g` are activation functions.
        - :math:`\\mathrm{\\beta}` is the *proximity* hyperparameter.
        - :math:`\\mathbf{O}` is a randomly generated orthogonal matrix.


    Parameters
    ----------
    units : int, optional
        Number of reservoir units. If None, the number of units will be inferred from
        the ``W`` matrix shape.
    proximity : float or array-like of shape (units,), default to 1.0
        Proximity parameter. Must be in :math:`[0, 1]`.
    sr : float, optional
        Spectral radius of recurrent weight matrix.
    input_scaling : float or array-like of shape (features,), default to 1.0.
        Input gain. An array of the same dimension as the inputs can be used to
        set up different input scaling for each feature.
    input_connectivity : float, default to 0.1
        Connectivity of input neurons, i.e. ratio of input neurons connected
        to reservoir neurons. Must be in :math:`]0, 1]`.
    rc_connectivity : float, default to 0.1
        Connectivity of recurrent weight matrix, i.e. ratio of reservoir
        neurons connected to other reservoir neurons, including themselves.
        Must be in :math:`]0, 1]`.
    Win : callable or array-like of shape (units, features), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Input weights matrix or initializer. If a callable (like a function) is used,
        then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    W : callable or array-like of shape (units, units), default to :py:func:`~reservoirpy.mat_gen.normal`
        Recurrent weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    O : callable or array-like of shape (units, units), default to :py:func:`~reservoirpy.mat_gen.orthogonal`
        Orthogonal matrix. If a callable (like a function) is used, then this function should
        accept any keywords parameters and at least two parameters that will be used to define
        the shape of the returned weight matrix.
    bias : callable or array-like of shape (units, 1), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    activation : str or callable, default to :py:func:`~reservoirpy.activationsfunc.tanh`
        Reservoir units activation function.
        - If a str, should be a :py:mod:`~reservoirpy.activationsfunc`
        function name.
        - If a callable, should be an element-wise operator on arrays.
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    dtype : Numpy dtype, default to np.float64
        Numerical type for node parameters.
    seed : int or :py:class:`numpy.random.Generator`, optional
        A random state seed, for noise generation.
    name : str, optional
        Node name.

    Note
    ----

    If W, Win, bias or Wfb are initialized with an array-like matrix, then all
    initializers parameters such as spectral radius (``sr``) or input scaling
    (``input_scaling``) are ignored.
    See :py:mod:`~reservoirpy.mat_gen` for more information.

    Example
    -------

    >>> from reservoirpy.nodes import ES2N, Ridge
    >>> es2n = ES2N(100, proximity=0.2, sr=0.8) >> Ridge(ridge=1e-6)

    Using the :py:func:`~reservoirpy.datasets.mackey_glass` timeseries:

    >>> from reservoirpy.datasets import mackey_glass, to_forecasting
    >>> x, y = to_forecasting(mackey_glass(200), forecast=10)
    >>> states = es2n.fit(x, y)

    .. plot::

        from reservoirpy.nodes import ES2N
        reservoir = ES2N(100, proximity=0.2, sr=0.8)
        from reservoirpy.datasets import mackey_glass
        x = mackey_glass(200)
        states = reservoir.run(x)
        fig, ax = plt.subplots(6, 1, figsize=(7, 10), sharex=True)
        ax[0].plot(x)
        ax[0].grid()
        ax[0].set_title("Neuron states (on Mackey-Glass)")
        for i in range(1, 6):
            ax[i].plot(states[:, i], label=f"Neuron {i}")
            ax[i].legend()
            ax[i].grid()
        ax[-1].set_xlabel("Timesteps")

    References
    ==========

        .. [1] Ceni, A., & Gallicchio, C. (2024). Edge of stability echo state network.
               IEEE Transactions on Neural Networks and Learning Systems, 36(4), 7555-7564.
    """

    initialized: bool
    input_dim: Optional[int]
    output_dim: int
    name: Optional[str]
    state: State

    # params
    #: Number of neuronal units in the reservoir.
    units: int
    #: Leaking rate (1.0 by default) (:math:`\mathrm{lr}`).
    proximity: Union[float, np.ndarray]
    #: Spectral radius of ``W`` (optional).
    sr: float
    #: Input scaling (float or array) (1.0 by default).
    input_scaling: Union[float, Sequence]
    #: Connectivity (or density) of ``Win`` (0.1 by default).
    input_connectivity: float
    #: Connectivity (or density) of ``Wfb`` (0.1 by default).
    rc_connectivity: float
    #: Input weights matrix (:math:`\mathbf{W}_{in}`).
    Win: Union["Weights", Callable]
    #: Recurrent weights matrix (:math:`\mathbf{W}`).
    W: Union["Weights", Callable]
    #: Orthogonal matrix (:math:`\mathbf{O}`).
    O: Union["Weights", Callable]
    #: Bias vector (:math:`\mathbf{b}`).
    bias: Union["Weights", Callable, float]
    #: Type of matrices elements. By default, ``np.float64``.
    dtype: type
    #: Activation of the reservoir units (tanh by default) (:math:`f`).
    activation: Callable
    #: A random state generator. Used for generating Win and W.
    rng: np.random.Generator

    def __init__(
        self,
        units: Optional[int] = None,
        proximity: Union[float, np.ndarray] = 1.0,
        sr: float = 1.0,
        input_scaling: Union[float, Sequence] = 1.0,
        input_connectivity: float = 0.1,
        rc_connectivity: float = 0.1,
        Win: Union["Weights", Callable] = bernoulli,
        W: Union["Weights", Callable] = normal,
        O: Union["Weights", Callable] = orthogonal,
        bias: Union["Weights", Callable, float] = 0.0,
        activation: Union[str, Callable] = tanh,
        input_dim: Optional[int] = None,
        dtype: type = np.float64,
        seed: Optional[Union[int, np.random.Generator]] = None,
        name: Optional[str] = None,
    ):

        self.proximity = proximity
        self.sr = sr
        self.input_scaling = input_scaling
        self.input_connectivity = input_connectivity
        self.rc_connectivity = rc_connectivity
        self.Win = Win
        self.W = W
        self.O = O
        self.bias = bias
        self.activation = get_function(activation)
        self.dtype = dtype
        self.rng = random.rand_generator(seed=seed)
        self.initialized = False
        self.name = name

        # set units / output_dim
        if units is None and not is_array(W):
            raise ValueError("'units' parameter must not be None if 'W' parameter is not a matrix.")
        if units is not None and is_array(W) and W.shape[-1] != units:
            raise ValueError(
                f"Both 'units' and 'W' are set but their dimensions doesn't match: " f"{units} != {W.shape[-1]}."
            )
        self.units = units if units is not None else W.shape[-1]
        self.output_dim = self.units

        # set input_dim (if possible)
        if input_dim is not None and is_array(Win) and Win.shape[-1] != input_dim:
            raise ValueError(
                f"Both 'input_dim' and 'Win' are set but their dimensions doesn't "
                f"match: {input_dim} != {Win.shape[-1]}."
            )
        self.input_dim = Win.shape[-1] if is_array(Win) else input_dim

    def initialize(self, x: Optional[Union[NodeInput, Timestep]]):

        # set input_dim
        self._set_input_dim(x)

        [Win_rng, W_rng, bias_rng, O_rng] = self.rng.spawn(4)

        if callable(self.Win):
            self.Win = self.Win(
                self.units,
                self.input_dim,
                input_scaling=self.input_scaling,
                connectivity=self.input_connectivity,
                dtype=self.dtype,
                seed=Win_rng,
            )

        if callable(self.W):
            self.W = self.W(
                self.units,
                self.units,
                sr=self.sr,
                connectivity=self.rc_connectivity,
                dtype=self.dtype,
                seed=W_rng,
            )

        if callable(self.O):
            self.O = self.O(
                self.units,
                self.units,
                dtype=self.dtype,
                seed=O_rng,
            )

        if callable(self.bias):
            self.bias = self.bias(
                self.units,
                connectivity=1.0,
                dtype=self.dtype,
                seed=bias_rng,
            )

        self.state = {"out": np.zeros((self.units,))}

        self.initialized = True

    def _step(self, state: State, x: Timestep) -> State:
        W = self.W  # NxN
        O = self.O  # NxN
        Win = self.Win  # NxI
        bias = self.bias  # N or float
        f = self.activation
        beta = self.proximity
        s = state["out"]

        next_state = f(W @ s + Win @ x + bias)
        next_state = (1 - beta) * O @ s + beta * next_state

        return {"out": next_state}
